"""Defines central controller for multiple expressions.

Michel Bierlaire
15.04.2025 09:43
"""

from __future__ import annotations

import logging
import random
from collections.abc import Iterator
from functools import reduce
from itertools import product
from typing import Any, Callable

import biogeme
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, ExpressionCollector, NamedExpression
from biogeme.parameters import get_default_value
from .catalog import Catalog
from .configuration import Configuration, SEPARATOR, SelectionTuple
from .controller import Controller

ControllerOperator = Callable[
    [Configuration, int],
    tuple[Configuration, int],
]

logger = logging.getLogger(__name__)


def extract_multiple_expressions_controllers(
    the_expression: Expression,
) -> list[Controller]:
    # Create walker
    walker = ExpressionCollector()

    @walker.register(Catalog)
    def retrieve_controllers(
        expr: Catalog, context: Any | None = None
    ) -> list[Controller]:
        return [expr.controlled_by]

    # Now use it
    return walker.walk(the_expression)


def count_number_of_specifications(expression: Expression):
    central_controller = CentralController(expression=expression)
    return central_controller.number_of_configurations()


class CentralController:
    """Class controlling the complete multiple expression"""

    def __init__(
        self,
        expression: Expression,
        maximum_number_of_configurations: int | None = None,
    ):
        """Constructor

        :param expression: controllers expression
        :type expression: biogeme.expressions.Expression

        """
        set_of_controllers = set(
            extract_multiple_expressions_controllers(the_expression=expression)
        )

        self.controllers: tuple[Controller, ...] = tuple(sorted(set_of_controllers))
        self.dict_of_controllers = {
            controller.controller_name: controller for controller in self.controllers
        }
        self.expression = expression

        all_controllers_states = [
            the_controller.all_configurations() for the_controller in self.controllers
        ]

        # The total number of configurations is the product of the
        # length of each controller.

        if all_controllers_states:
            self._number_of_configurations = reduce(
                lambda x, y: x * y, map(len, all_controllers_states)
            )
        else:
            self._number_of_configurations = 0

        if maximum_number_of_configurations is None:
            maximum_number_of_configurations = get_default_value(
                name="maximum_number_catalog_expressions", section="Estimation"
            )
        if self.number_of_configurations() > maximum_number_of_configurations:
            self.all_configurations_ids = None
            self.all_configurations = None
            return

        self.all_configurations_ids = {
            SEPARATOR.join(combination)
            for combination in product(*all_controllers_states)
        }
        self.all_configurations = {
            Configuration.from_string(conf_id)
            for conf_id in self.all_configurations_ids
        }

    def reset_selection(self) -> None:
        for controller in self.controllers:
            controller.reset_selection()

    def number_of_configurations(self) -> int:
        """Total number of configurations"""
        return self._number_of_configurations

    def get_configuration(self) -> Configuration:
        """Obtain the current configuration of the controllers"""
        selections = (
            SelectionTuple(
                controller=controller.controller_name,
                selection=controller.current_name(),
            )
            for controller in self.controllers
        )
        return Configuration(selections)

    def set_configuration(self, configuration: Configuration) -> None:
        """Apply a configuration to the controllers

        :param configuration: the configuration to be applied
        :type configuration: Configuration
        """
        properly_set = {
            controller.controller_name: False for controller in self.controllers
        }
        for selection in configuration.selections:
            controller: Controller = self.dict_of_controllers.get(selection.controller)
            if controller is None:
                error_msg = (
                    f"Wrong configuration: {configuration}. Controller "
                    f"{selection.controller} is unknown"
                )
                raise BiogemeError(error_msg)
            controller.set_name(selection.selection)
            properly_set[selection.controller] = True

        missing_controllers = [name for name, done in properly_set.items() if not done]
        if missing_controllers:
            error_msg = (
                f"Incomplete configuration {configuration}. The following controllers "
                f"are not defined: {missing_controllers}"
            )
            raise BiogemeError(error_msg)

    def set_configuration_from_id(self, configuration_id: str) -> None:
        """Apply a configuration to the controllers

        :param configuration_id: the ID of the configuration to be applied
        :type configuration_id: str
        """
        the_configuration = Configuration.from_string(configuration_id)
        self.set_configuration(the_configuration)

    def set_controller(self, controller_name: str, index: int) -> None:
        """Set the given controller to the specified index

        :param controller_name: name of the controller
        :type controller_name: string

        :param index: index of the selection
        :type index: int
        """
        the_controller = self.dict_of_controllers.get(controller_name)

        if the_controller is None:
            error_msg = f'Unknown controller: {controller_name}'
            raise BiogemeError(error_msg)

        the_controller.set_index(index)

    def expression_iterator(self) -> Iterator[NamedExpression]:
        """Iterate over the expression corresponding to each configuration."""
        if self.all_configurations is None:
            return
        for config in self.all_configurations:
            self.set_configuration(config)
            yield NamedExpression(name=config.string_id, expression=self.expression)

    def expression_configuration_iterator(self) -> Iterator[str]:
        """Iterate over the expression corresponding to each configuration."""
        if self.all_configurations is None:
            return
        for config in self.all_configurations:
            self.set_configuration(config)
            yield config

    def increased_controller(
        self,
        controller_name: str,
        current_config: Configuration,
        step: int,
    ) -> tuple[Configuration, int]:
        """Increase the selection of one controller by "step"

        :param controller_name: name of the controller to modify
        :type controller_name: str

        :param current_config: current configuration
        :type current_config: str

        :param step: number of steps to perform
        :type step: int
        """
        self.set_configuration(current_config)
        the_controller = self.dict_of_controllers.get(controller_name)
        if the_controller is None:
            error_msg = f'Unknown controller {the_controller}'
            raise BiogemeError(error_msg)
        the_controller.modify_controller(step=step, circular=True)
        new_config = self.get_configuration()
        return new_config, step

    def get_operator_increased_controllers(
        self, controller_name: str
    ) -> ControllerOperator:

        def operator_increased_controller(
            current_config: Configuration, step: int
        ) -> tuple[Configuration, int]:
            return self.increased_controller(
                controller_name=controller_name,
                current_config=current_config,
                step=step,
            )

        return operator_increased_controller

    def decreased_controller(
        self,
        controller_name: str,
        current_config: Configuration,
        step: int,
    ) -> tuple[Configuration, int]:
        """Decrease the selection of one controller by "step"

        :param controller_name: name of the controller to modify
        :param current_config: current configuration
        :param step: number of steps to perform
        :return: ID of the new configuration and number of steps performed.
        """
        self.set_configuration(current_config)
        the_controller = self.dict_of_controllers.get(controller_name)
        if the_controller is None:
            error_msg = f'Unknown controller {the_controller}'
            raise BiogemeError(error_msg)
        the_controller.modify_controller(step=-step, circular=True)
        new_config = self.get_configuration()
        return new_config, step

    def get_operator_decreased_controllers(
        self, controller_name: str
    ) -> ControllerOperator:

        def operator_decreased_controller(
            current_config: Configuration, step: int
        ) -> tuple[Configuration, int]:
            return self.decreased_controller(
                controller_name=controller_name,
                current_config=current_config,
                step=step,
            )

        return operator_decreased_controller

    def two_controllers(
        self,
        first_controller_name: str,
        second_controller_name: str,
        direction: str,
        current_config: Configuration,
        step: int,
    ) -> tuple[Configuration, int]:
        """Modification of two controllers. Meaning of direction:

        - NE (North-East): first controller increased by "step", second
              controller increased by "step"
        - NW (North-West): first controller decreased by "step", second
              controller increased by "step"
        - SE (South-East): first controller increased by "step", second
              controller decreased by "step"
        - SW (South-West): first controller decreased by "step", second
              controller decreased by "step"

        :param first_controller_name: name of the first_controller to modify
        :type first_controller_name: str

        :param second_controller_name: name of the second_controller to modify
        :param direction: direction based on the compass rose. Must be NE, NW, SE or SW
        :param current_config: current configuration
        :param step: number of steps to perform
        :return: ID of the new configuration and number of steps performed.

        """
        if direction not in ["NE", "NW", "SE", "SW"]:
            error_msg = f'Incorrect direction {direction}. Must be one of {direction}'
            raise BiogemeError(error_msg)
        self.set_configuration(current_config)
        the_first_controller = self.dict_of_controllers.get(first_controller_name)
        if the_first_controller is None:
            error_msg = f'Unknown controller {the_first_controller}'
            raise BiogemeError(error_msg)
        the_second_controller = self.dict_of_controllers.get(second_controller_name)
        if the_second_controller is None:
            error_msg = f'Unknown controller {the_second_controller}'
            raise BiogemeError(error_msg)
        # The direction for the first controller is W (decrease) or E (increase)
        # The direction for the second controller is N (increase) or S (decrease)
        actual_step = step if direction[1] == "E" else -step
        the_first_controller.modify_controller(step=actual_step, circular=True)
        actual_step = step if direction[0] == "N" else -step
        the_second_controller.modify_controller(step=actual_step, circular=True)
        new_config = self.get_configuration()
        return new_config, step

    def get_operator_two_controllers(
        self, first_controller_name: str, second_controller_name: str, direction: str
    ) -> ControllerOperator:

        def operator_two_controller(
            current_config: Configuration, step: int
        ) -> tuple[Configuration, int]:
            return self.two_controllers(
                first_controller_name=first_controller_name,
                second_controller_name=second_controller_name,
                direction=direction,
                current_config=current_config,
                step=step,
            )

        return operator_two_controller

    def modify_random_controllers(
        self,
        increase: bool,
        current_config: Configuration,
        step: int,
    ) -> tuple[Configuration, int]:
        """Increase the selection of "step" controllers by 1

        :param increase: If True, the indices are increased . If
            False, they are decreased.
        :param current_config: current configuration
        :param step: number of steps to perform

        """
        self.set_configuration(current_config)
        number_of_controllers = len(self.controllers)
        actual_size = min(step, number_of_controllers)
        selected_controllers = random.choices(
            list(self.dict_of_controllers.keys()), k=actual_size
        )
        the_modification = 1 if increase else 1
        for the_controller in selected_controllers:
            self.dict_of_controllers[the_controller].modify_controller(
                step=the_modification, circular=True
            )
        new_config = self.get_configuration()
        return new_config, actual_size

    def get_operator_modify_random_controllers(
        self, increase: bool
    ) -> ControllerOperator:

        def operator_modify_random_controller(
            current_config: Configuration, step: int
        ) -> tuple[Configuration, int]:
            return self.modify_random_controllers(
                increase=increase,
                current_config=current_config,
                step=step,
            )

        return operator_modify_random_controller

    def prepare_operators(
        self,
    ) -> dict[str, ControllerOperator]:
        """Operators are functions that take a configuration and a
        size as arguments, and return a new configuration, and the
        actual number of modifications that have been
        implemented.

        """

        dict_of_operators = {}
        # Increase and decrease controllers
        for name in self.dict_of_controllers.keys():
            dict_of_operators[f"Increase {name}"] = (
                self.get_operator_increased_controllers(controller_name=name)
            )
            dict_of_operators[f"Decrease {name}"] = (
                self.get_operator_decreased_controllers(controller_name=name)
            )
        # Pair of controllers
        directions = ["NE", "NW", "SE", "SW"]
        # Using nested list comprehensions to generate all possible pairs
        all_pairs = [
            (name1, name2, direction)
            for name1 in self.dict_of_controllers.keys()
            for name2 in self.dict_of_controllers.keys()
            for direction in directions
        ]

        # Filter out pairs with the same name
        filtered_pairs = [
            (name1, name2, direction)
            for name1, name2, direction in all_pairs
            if name1 != name2
        ]

        for name1, name2, direction in filtered_pairs:
            dict_of_operators[f"Pair_{name1}_{name2}_{direction}"] = (
                self.get_operator_two_controllers(
                    first_controller_name=name1,
                    second_controller_name=name2,
                    direction=direction,
                )
            )

        # Several controllers
        dict_of_operators["Increase_several"] = (
            self.get_operator_modify_random_controllers(increase=True)
        )
        dict_of_operators["Decrease_several"] = (
            self.get_operator_modify_random_controllers(increase=False)
        )
        return dict_of_operators



""".. _plot_b22process_pareto:

Re-estimate the Pareto optimal models
=====================================

The assisted specification algorithm generates a file containg the
pareto optimal specification. This script is designed to re-estimate
the Pareto optimal models. The catalog of specifications is defined in
:ref:`plot_b22multiple_models_spec` .

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 17:25:41 2023


"""
try:
    import matplotlib.pyplot as plt

    can_plot = True
except ModuleNotFoundError:
    can_plot = False
from biogeme.assisted import ParetoPostProcessing
from biogeme.results import compileEstimationResults
from plot_b22multiple_models_spec import the_biogeme

PARETO_FILE_NAME = 'saved_results/b22multiple_models.pareto'

CSV_FILE = 'b22process_pareto.csv'
SEP_CSV = ','

# %%
# The constructor of the Pareto post processing object takes two arguments:
#
#    - the biogeme object,
#    - the name of the file where the algorithm has stored the
#      estimated models.
the_pareto_post = ParetoPostProcessing(
    biogeme_object=the_biogeme,
    pareto_file_name=PARETO_FILE_NAME,
)

# %%
the_pareto_post.log_statistics()

# %%
all_results = the_pareto_post.reestimate(recycle=True)

# %%
summary, description = compileEstimationResults(all_results, use_short_names=True)
print(summary)

# %%
print(f'Summary table available in {CSV_FILE}')
summary.to_csv(CSV_FILE, sep=SEP_CSV)

# %%
# Explanation of the short names of the models.
with open(CSV_FILE, 'a', encoding='utf-8') as f:
    print('\n\n', file=f)
    for k, v in description.items():
        if k != v:
            print(f'{k}: {v}')
            print(f'{k}{SEP_CSV}{v}', file=f)


# %%
# The following plot illustrates all models that have been
# estimated.  Each dot corresponds to a model. The x-coordinate
# corresponds to the Akaike Information Criterion (AIC). The
# y-coordinate corresponds to the Bayesian Information Criterion
# (BIC). Note that there is a third objective that does not appear on
# this picture: the number of parameters. If the shape of the dot is a
# circle, it means that it corresponds to a Pareto optimal model. If
# the shape is a cross, it means that the model has been Pareto
# optimal at some point during the algorithm and later removed as a
# new model dominating it has been found. If the shape is a start, it
# means that the model has been deemed invalid.
if can_plot:
    _ = the_pareto_post.plot(label_x='AIC', label_y='BIC')
    plt.show()

# %%
# It is possible to plot two different objectives: AIC and number of parameters.
if can_plot:
    _ = the_pareto_post.plot(
        objective_x=0, objective_y=2, label_x='AIC', label_y='Number of parameters'
    )
    plt.show()

# %%
# It is possible to plot two different objectives: BIC and number of parameters.
if can_plot:
    _ = the_pareto_post.plot(
        objective_x=1, objective_y=2, label_x='BIC', label_y='Number of parameters'
    )
    plt.show()



"""

Antithetic draws explicitly generated
=====================================

Calculation of a simple integral using Monte-Carlo integration. It
illustrates how to use antothetic draws, explicitly generared.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 20:49:50 2023
"""

import numpy as np
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import draws
from biogeme.expressions import exp, bioDraws, MonteCarlo

# %%
R = 10000

# %%
# We create a fake database with one entry, as it is required to store
# the draws
df = pd.DataFrame()
df['FakeColumn'] = [1.0]
database = db.Database('fake_database', df)


# %%
def halton13(sample_size: int, number_of_draws: int) -> np.array:
    """
    The user can define new draws. For example, Halton draws
    with base 13, skipping the first 10 draws.

    :param sample_size: number of observations for which draws must be
                       generated.
    :param number_of_draws: number of draws to generate.

    """
    return draws.getHaltonDraws(sample_size, number_of_draws, base=13, skip=10)


# %%
my_draws = {'HALTON13': (halton13, 'Halton draws, base 13, skipping 10')}
database.setRandomNumberGenerators(my_draws)

# %%
U = bioDraws('U', 'UNIFORM')
integrand = exp(U) + exp(1 - U)
simulatedI = MonteCarlo(integrand) / 2.0

# %%
U_halton13 = bioDraws('U_halton13', 'HALTON13')
integrand_halton13 = exp(U_halton13) + exp(1 - U_halton13)
simulatedI_halton13 = MonteCarlo(integrand_halton13) / 2.0

# %%
U_mlhs = bioDraws('U_mlhs', 'UNIFORM_MLHS')
integrand_mlhs = exp(U_mlhs) + exp(1 - U_mlhs)
simulatedI_mlhs = MonteCarlo(integrand_mlhs) / 2.0

# %%
trueI = exp(1.0) - 1.0

# %%
error = simulatedI - trueI

# %%
error_halton13 = simulatedI_halton13 - trueI

# %%
error_mlhs = simulatedI_mlhs - trueI

# %%
simulate = {
    'Analytical Integral': trueI,
    'Simulated Integral': simulatedI,
    'Error             ': error,
    'Simulated Integral (Halton13)': simulatedI_halton13,
    'Error (Halton13)             ': error_halton13,
    'Simulated Integral (MLHS)': simulatedI_mlhs,
    'Error (MLHS)             ': error_mlhs,
}

# %%
biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b03antithetic_explicit'

# %%
results = biosim.simulate(theBetaValues={})
results

# %%
print(f"Analytical integral: {results.iloc[0]['Analytical Integral']:.6f}")
print(
    f"         \t{'Uniform (Anti)':>15}\t{'Halton13 (Anti)':>15}\t{'MLHS (Anti)':>15}"
)
print(
    f"Simulated\t{results.iloc[0]['Simulated Integral']:>15.6g}\t"
    f"{results.iloc[0]['Simulated Integral (Halton13)']:>15.6g}\t"
    f"{results.iloc[0]['Simulated Integral (MLHS)']:>15.6g}"
)
print(
    f"Error\t\t{results.iloc[0]['Error             ']:>15.6g}\t"
    f"{results.iloc[0]['Error (Halton13)             ']:>15.6g}\t"
    f"{results.iloc[0]['Error (MLHS)             ']:>15.6g}"
)



import logging

import numpy as np

from biogeme.exceptions import BiogemeError

logger = logging.getLogger(__name__)


def get_prime_numbers(n: int) -> list[int]:
    """Get a given number of prime numbers.

    :param n: Number of primes that are requested.
    :return: List with prime numbers.
    :raise BiogemeError: If the requested number is non-positive.
    """
    if not isinstance(n, int) or n <= 0:
        raise BiogemeError(f'Incorrect number: {n}')

    # Initial setup based on known distribution of primes for efficiency
    if n == 1:
        return [2]  # The first prime number is 2
    estimated_upper_bound = n * (
        np.log(n) + np.log(np.log(n))
    )  # Estimation based on the prime number theorem
    primes = calculate_prime_numbers(int(estimated_upper_bound))

    # In case the estimation was too low, which is very rare, increase until enough primes are found
    while len(primes) < n:
        estimated_upper_bound *= 1.5
        primes = calculate_prime_numbers(int(estimated_upper_bound))

    return primes[:n]


def calculate_prime_numbers(upper_bound: int) -> list[int]:
    """Calculate prime numbers up to a specified upper bound using the Sieve of Eratosthenes.

    :param upper_bound: Prime numbers up to this value will be computed.
    :return: A list of prime numbers up to the upper bound.
    :raise BiogemeError: If the sqrt_max is incorrectly defined (e.g., negative number).
    """
    if not isinstance(upper_bound, int) or upper_bound < 0:
        raise BiogemeError(f'Incorrect value: {upper_bound}')
    if upper_bound < 2:
        return []  # There are no prime numbers less than 2

    # Initialize a list indicating whether numbers are prime
    is_prime = [False, False] + [True] * (
        upper_bound - 1
    )  # 0 and 1 are not prime, others assumed prime initially
    for number in range(
        2, int(np.sqrt(upper_bound)) + 1
    ):  # Check up to the square root of the sqrt_max
        if is_prime[number]:  # Found a prime
            for multiple in range(
                number * number, upper_bound + 1, number
            ):  # Mark multiples of the prime as non-prime
                is_prime[multiple] = False

    # Extract prime numbers based on the boolean list
    primes = [num for num, prime in enumerate(is_prime) if prime]

    return primes



"""
Test the estimation results

Michel Bierlaire
Fri Oct 4 09:52:23 2024
"""

import unittest
import numpy as np
from datetime import timedelta
from unittest.mock import patch, MagicMock

from biogeme.exceptions import BiogemeError
from biogeme.results_processing import (
    EstimationResults,
    RawEstimationResults,
    calc_p_value,
    EstimateVarianceCovariance,
    get_pandas_estimated_parameters,
)


class TestEstimationResults(unittest.TestCase):

    def setUp(self):
        """Set up the test fixtures."""
        self.raw_results = RawEstimationResults(
            model_name='Test Model',
            user_notes='Test notes',
            beta_names=['beta1', 'beta2'],
            beta_values=[1.0, 1.5],
            lower_bounds=[0.0, 1.5],
            upper_bounds=[1.5, 2.5],
            gradient=[0.1, 0.2],
            hessian=[[-0.5, -0.2], [-0.2, -0.4]],
            bhhh=[[0.01, 0.02], [0.02, 0.03]],
            null_log_likelihood=-300.0,
            initial_log_likelihood=-200.0,
            final_log_likelihood=-150.0,
            data_name='Sample Data',
            sample_size=1000,
            number_of_observations=100,
            monte_carlo=False,
            number_of_draws=500,
            types_of_draws={'type1': 'UNIFORM'},
            number_of_excluded_data=10,
            draws_processing_time=timedelta(seconds=300),
            optimization_messages={'message': 'Optimization successful'},
            convergence=True,
            bootstrap=[[1.0, 2.0], [1.1, 2.1]],
            bootstrap_time=timedelta(seconds=100),
        )
        self.estimation_results = EstimationResults(self.raw_results)

    def test_empty(self):
        with self.assertRaises(BiogemeError):
            the_empty = EstimationResults(raw_estimation_results=None)

    def test_number_of_parameters(self):
        """Test the number_of_parameters property."""
        self.assertEqual(self.estimation_results.number_of_parameters, 2)

    def test_number_of_free_parameters(self):
        """Test the number_of_free_parameters property."""
        self.assertEqual(self.estimation_results.number_of_free_parameters, 1)

    def test_is_any_bound_active(self):
        """Test if any bound is active."""
        self.assertTrue(self.estimation_results.is_any_bound_active())

    def test_is_bound_active(self):
        """Test if a specific bound is active."""
        self.assertFalse(self.estimation_results.is_bound_active('beta1'))
        self.assertTrue(self.estimation_results.is_bound_active('beta2'))

    def test_akaike_information_criterion(self):
        """Test AIC calculation."""
        expected_aic = 2.0 * 2 - 2 * -150.0
        self.assertEqual(
            self.estimation_results.akaike_information_criterion, expected_aic
        )

    def test_bayesian_information_criterion(self):
        """Test BIC calculation."""
        expected_bic = -2.0 * -150.0 + 2 * np.log(1000)
        self.assertAlmostEqual(
            self.estimation_results.bayesian_information_criterion, expected_bic
        )

    def test_algorithm_has_converged(self):
        """Test if the algorithm has converged."""
        self.assertTrue(self.estimation_results.algorithm_has_converged)

    def test_variance_covariance_missing(self):
        """Test if the variance covariance matrix is missing."""
        self.assertFalse(self.estimation_results.variance_covariance_missing)

    def test_eigen_structure(self):
        """Test eigen structure calculation."""
        eigenvalues, eigenvectors = self.estimation_results.eigen_structure()
        self.assertEqual(len(eigenvalues), 2)
        self.assertEqual(eigenvectors.shape, (2, 2))

    def test_smallest_eigenvalue(self):
        """Test smallest eigenvalue calculation."""
        self.assertAlmostEqual(
            self.estimation_results.smallest_eigenvalue, 0.24384471871911698, places=7
        )

    def test_largest_eigenvalue(self):
        """Test largest eigenvalue calculation."""
        self.assertAlmostEqual(
            self.estimation_results.largest_eigenvalue, 0.6561552812808831, places=7
        )

    def test_condition_number(self):
        """Test condition number calculation."""
        expected_condition_number = (
            self.estimation_results.largest_eigenvalue
            / self.estimation_results.smallest_eigenvalue
        )
        self.assertAlmostEqual(
            self.estimation_results.condition_number, expected_condition_number
        )

    def test_rao_cramer_variance_covariance_matrix(self):
        """Test Rao-Cramer variance-covariance matrix calculation."""
        rao_cramer_matrix = (
            self.estimation_results.rao_cramer_variance_covariance_matrix
        )
        self.assertEqual(rao_cramer_matrix.shape, (2, 2))

    def test_likelihood_ratio_null(self):
        """Test likelihood ratio test against the null model."""
        expected_lr_null = -2 * (-300.0 - (-150.0))
        self.assertEqual(
            self.estimation_results.likelihood_ratio_null, expected_lr_null
        )

    def test_get_parameter_value(self):
        """Test parameter value retrieval."""
        self.assertEqual(self.estimation_results.get_parameter_value_from_index(1), 1.5)

    def test_get_parameter_t_test(self):
        """Test t-test calculation for a parameter."""
        t_value = self.estimation_results.get_parameter_t_test_from_index(
            parameter_index=0, estimate_var_covar=EstimateVarianceCovariance.RAO_CRAMER
        )
        self.assertAlmostEqual(np.abs(t_value), 0.632455532033676)
        t_value = self.estimation_results.get_parameter_t_test_from_index(
            parameter_index=0,
            estimate_var_covar=EstimateVarianceCovariance.RAO_CRAMER,
            target=2.0,
        )
        self.assertAlmostEqual(np.abs(t_value), 0.632455532033676)

    def test_get_parameter_p_value(self):
        """Test p-value calculation for a parameter."""
        p_value = self.estimation_results.get_parameter_p_value_from_index(
            0, EstimateVarianceCovariance.RAO_CRAMER
        )
        self.assertAlmostEqual(p_value, 0.5270892568655381)

    def test_calc_p_value(self):
        """Test the p-value calculation using calc_p_value function."""
        t_stat = 1.96
        p_value = calc_p_value(t_stat)
        self.assertAlmostEqual(p_value, 0.05, places=2)

    def test_likelihood_ratio_test(self):
        """Test likelihood ratio test between two models."""
        other_model = MagicMock(spec=EstimationResults)
        other_model.final_log_likelihood = -140.0
        other_model.number_of_parameters = 3

        # Mock the likelihood ratio function
        with patch(
            'biogeme.tools.likelihood_ratio.likelihood_ratio_test',
            return_value=("Pass", 2.0, 3.84),
        ):
            result = self.estimation_results.likelihood_ratio_test(other_model)
            self.assertEqual(result, ("Pass", 2.0, 3.84))

    def test_get_beta_values(self):
        """Test retrieval of beta values."""
        betas = self.estimation_results.get_beta_values()
        self.assertEqual(betas['beta1'], 1.0)

    def test_bootstrap_variance_covariance_matrix(self):
        """Test bootstrap variance-covariance matrix."""
        bootstrap_var_covar = (
            self.estimation_results.bootstrap_variance_covariance_matrix
        )
        self.assertEqual(bootstrap_var_covar.shape, (2, 2))

    def test_short_summary(self):
        """Test the short summary method."""
        summary = self.estimation_results.short_summary()
        self.assertIn('Test Model', summary)
        self.assertIn('Nbr of parameters:', summary)
        self.assertIn('Sample size:', summary)

    def test_get_general_statistics(self):
        """Test the general statistics dictionary generation."""
        stats = self.estimation_results.get_general_statistics()
        self.assertIn('Number of estimated parameters', stats)
        self.assertIn('Sample size', stats)

    def test_dump_yaml_file(self):
        """Test the YAML dumping of raw estimation results."""
        with patch(
            'biogeme.results_processing.estimation_results.serialize_to_yaml'
        ) as mock_serialize:
            self.estimation_results.dump_yaml_file('test.yaml')
            mock_serialize.assert_called_once_with(
                data=self.raw_results, filename='test.yaml'
            )

    def test_pandas_results(self):
        pandas_results = get_pandas_estimated_parameters(
            estimation_results=self.estimation_results
        )
        nrows, ncolumns = pandas_results.shape
        self.assertEqual(nrows, 2)
        self.assertEqual(ncolumns, 6)


if __name__ == '__main__':
    unittest.main()



"""
Obtain sampling weights using local-sensitivity hashing

:author: Nicola Ortelli
:date: Fri Aug 11 18:25:39 2023

"""

import numpy as np
import pandas as pd


def get_lsh_weights(
    df: pd.DataFrame, w: float, a: np.ndarray, max_weight: int | None
) -> np.ndarray:
    """Compute weights using Locality-Sensitive Hashing (LSH) on input data.

    This function applies LSH to the input data frame, generating weights
    based on bucketing of the data. It also provides an option to limit
    the maximum weight assigned to a group of data points.

    :param df: The input data frame containing the data to compute weights for.
        The DataFrame should have at least one target column and one
        weight column.
    :type df: pandas DataFrame

    :param w: The width of the LSH buckets.
    :type w: float

    :param a: The LSH hash functions as a 2D array. Each row of this array
        represents an LSH hash function.
    :type a: numpy.ndarray

    :param max_weight: The maximum weight allowed for a group of data points. If not
        provided, no maximum weight constraint is applied.
    :type max_weight: int, optional

    :return: An array of weights corresponding to the input data frame.
    :rtype: numpy.ndarray
    """

    # Normalize the explanatory variables
    df_expla = df.drop(
        columns=['target', 'weight']
    )  # need to drop the intercept column...
    df_expla_norm = (df_expla - df_expla.min()) / (df_expla.max() - df_expla.min())

    # hashing into buckets according to LSH
    b = np.random.rand(a.shape[1]) * w
    buckets = np.floor((df_expla_norm.dot(a) + b) / w).astype(int)

    # if buckets depend on the target, uncomment this line...
    # buckets['target'] = df['target']

    # saving names of colummns storing buckets
    groupby_cols = list(buckets.columns)

    # Randomize the order of buckets to avoid bias
    buckets = buckets.sample(frac=1).reset_index()

    # adding a column that guarantees max_weight is never exceeded
    if max_weight:
        group_counts = buckets.groupby(groupby_cols, sort=False)[0].transform(
            'cumcount'
        )

        buckets['sieve'] = (group_counts / max_weight).astype(int)

        groupby_cols.append('sieve')

    # adding a column to store weights
    buckets['weight'] = 1

    # preparing aggregation dictionary for final grouping
    agg_dict = {'index': 'first', 'weight': 'sum'}

    # final grouping
    selected = buckets.groupby(groupby_cols, sort=False).agg(agg_dict)

    # building vector of weights
    weights = np.zeros(len(df))
    np.put(weights, selected['index'], selected['weight'])

    return weights



import copy
import unittest

from biogeme.expressions import (
    MultipleSum,
    PanelLikelihoodTrajectory,
    Variable,
    add_prefix_suffix_to_all_variables,
)


class TestDeepCopyInPanelLikelihoodTrajectory(unittest.TestCase):
    def test_deepcopy_children(self):
        # Original expression tree
        original_var = Variable('x')
        expression = MultipleSum([original_var, original_var + 1])
        panel_expr = PanelLikelihoodTrajectory(expression)
        panel_expr.set_maximum_number_of_observations_per_individual(3)

        # Generate copies
        copies = [
            copy.deepcopy(panel_expr.child)
            for _ in range(panel_expr.maximum_number_of_observations_per_individual)
        ]

        # Ensure copies are not the same object
        self.assertTrue(all(c is not panel_expr.child for c in copies))

        # Ensure modifying one copy does not affect the others
        number_of_modifications = add_prefix_suffix_to_all_variables(
            expr=copies[0], prefix='', suffix='_mod'
        )

        self.assertNotEqual(str(copies[0]), str(copies[1]))
        self.assertNotEqual(str(copies[0]), str(panel_expr.child))


if __name__ == '__main__':
    unittest.main()



import unittest

from biogeme.expressions import Variable
from biogeme.catalog import (
    Catalog,
    Controller,
    extract_multiple_expressions_controllers,
)


class TestExtractMultipleExpressions(unittest.TestCase):
    def test_no_catalog(self):
        """Test with a simple variable expression (no Catalog)"""
        expr = Variable('x')
        result = extract_multiple_expressions_controllers(expr)
        self.assertEqual(result, [])

    def test_one_catalog(self):
        """Test with one catalog expression"""
        expr1 = Variable('x')
        expr2 = Variable('y')
        controller = Controller(controller_name='ctrl', specification_names=['x', 'y'])
        expressions = {
            'x': expr1,
            'y': expr2,
        }
        catalog = Catalog.from_dict(
            catalog_name='C1', dict_of_expressions=expressions, controlled_by=controller
        )

        result = extract_multiple_expressions_controllers(catalog)
        self.assertEqual(len(result), 1)
        self.assertIs(result[0], controller)

    def test_nested_catalog(self):
        """Test with a nested catalog inside another"""
        expr1 = Variable('x')
        expr2 = Variable('y')
        controller_1 = Controller(
            controller_name='ctrl_1', specification_names=['x', 'y']
        )
        expressions_1 = {
            'x': expr1,
            'y': expr2,
        }
        catalog_1 = Catalog.from_dict(
            catalog_name='C1',
            dict_of_expressions=expressions_1,
            controlled_by=controller_1,
        )

        expr3 = Variable('z')
        expressions_2 = {
            'z': expr3,
            'c1': catalog_1,
        }
        controller_2 = Controller(
            controller_name='ctrl_2', specification_names=['z', 'c1']
        )
        catalog_2 = Catalog.from_dict(
            catalog_name='C2',
            dict_of_expressions=expressions_2,
            controlled_by=controller_2,
        )

        result = extract_multiple_expressions_controllers(catalog_2)
        self.assertEqual(set(result), {controller_1, controller_2})

    def test_duplicate_catalog(self):
        """Test that duplicate catalogs are returned separately"""
        expr1 = Variable('x')
        expr2 = Variable('y')
        controller = Controller(controller_name='ctrl', specification_names=['x', 'y'])
        expressions = {
            'x': expr1,
            'y': expr2,
        }
        catalog = Catalog.from_dict(
            catalog_name='C1', dict_of_expressions=expressions, controlled_by=controller
        )
        composite_expr = catalog + catalog

        result = extract_multiple_expressions_controllers(composite_expr)
        self.assertEqual(result.count(controller), 2)


if __name__ == "__main__":
    unittest.main()



"""
Test the draws module

:author: Michel Bierlaire
:data: Wed Apr 29 17:56:55 2020
"""

# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name, too-many-instance-attributes
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring

import unittest
import biogeme.draws as dr
import numpy as np
import biogeme.exceptions as excep


class test_draws(unittest.TestCase):
    def setUp(self):
        np.random.seed(90267)

    def test_getUniform(self):
        draws = dr.get_uniform(sample_size=5, number_of_draws=10, symmetric=False)
        r, c = draws.shape
        self.assertEqual(r, 5)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > 0)
        self.assertTrue(np.max(draws) <= 1)

    def test_getUniformSymmetric(self):
        draws = dr.get_uniform(sample_size=5, number_of_draws=10, symmetric=True)
        r, c = draws.shape
        self.assertEqual(r, 5)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > -1)
        self.assertTrue(np.max(draws) <= 1)

    def test_getUniform_exception(self):
        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_uniform(sample_size=5, number_of_draws=-10, symmetric=True)
        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_uniform(sample_size=-5, number_of_draws=10, symmetric=True)

    def test_getLatinHypercubeDraws(self):
        draws = dr.get_latin_hypercube_draws(sample_size=5, number_of_draws=10)
        r, c = draws.shape
        self.assertEqual(r, 5)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > 0)
        self.assertTrue(np.max(draws) <= 1)

    def test_getLatinHypercubeDrawsSymmetric(self):
        draws = dr.get_latin_hypercube_draws(
            sample_size=5, number_of_draws=10, symmetric=True
        )
        r, c = draws.shape
        self.assertEqual(r, 5)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > -1)
        self.assertTrue(np.max(draws) <= 1)

    def test_getLatinHypercubeDraws_exception(self):
        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_latin_hypercube_draws(sample_size=-5, number_of_draws=10)
        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_latin_hypercube_draws(sample_size=5, number_of_draws=-10)

        uniform_numbers = np.random.uniform(size=2)
        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_latin_hypercube_draws(
                sample_size=5, number_of_draws=10, uniform_numbers=uniform_numbers
            )

    def test_userDefinedDraws(self):
        myUnif = np.random.uniform(size=30)
        draws = dr.get_latin_hypercube_draws(
            sample_size=3,
            number_of_draws=10,
            symmetric=False,
            uniform_numbers=myUnif,
        )
        r, c = draws.shape
        self.assertEqual(r, 3)
        self.assertEqual(c, 10)
        self.assertTrue(np.min(draws) > 0)
        self.assertTrue(np.max(draws) <= 1)

    def test_halton(self):
        halton = dr.get_halton_draws(sample_size=1, number_of_draws=100, base=3)
        h = np.array(
            [
                0.33333333,
                0.66666667,
                0.11111111,
                0.44444444,
                0.77777778,
                0.22222222,
                0.55555556,
                0.88888889,
                0.03703704,
                0.37037037,
                0.7037037,
                0.14814815,
                0.48148148,
                0.81481481,
                0.25925926,
                0.59259259,
                0.92592593,
                0.07407407,
                0.40740741,
                0.74074074,
                0.18518519,
                0.51851852,
                0.85185185,
                0.2962963,
                0.62962963,
                0.96296296,
                0.01234568,
                0.34567901,
                0.67901235,
                0.12345679,
                0.45679012,
                0.79012346,
                0.2345679,
                0.56790123,
                0.90123457,
                0.04938272,
                0.38271605,
                0.71604938,
                0.16049383,
                0.49382716,
                0.82716049,
                0.27160494,
                0.60493827,
                0.9382716,
                0.08641975,
                0.41975309,
                0.75308642,
                0.19753086,
                0.5308642,
                0.86419753,
                0.30864198,
                0.64197531,
                0.97530864,
                0.02469136,
                0.35802469,
                0.69135802,
                0.13580247,
                0.4691358,
                0.80246914,
                0.24691358,
                0.58024691,
                0.91358025,
                0.0617284,
                0.39506173,
                0.72839506,
                0.17283951,
                0.50617284,
                0.83950617,
                0.28395062,
                0.61728395,
                0.95061728,
                0.09876543,
                0.43209877,
                0.7654321,
                0.20987654,
                0.54320988,
                0.87654321,
                0.32098765,
                0.65432099,
                0.98765432,
                0.00411523,
                0.33744856,
                0.67078189,
                0.11522634,
                0.44855967,
                0.781893,
                0.22633745,
                0.55967078,
                0.89300412,
                0.04115226,
                0.3744856,
                0.70781893,
                0.15226337,
                0.48559671,
                0.81893004,
                0.26337449,
                0.59670782,
                0.93004115,
                0.0781893,
                0.41152263,
            ]
        )
        diff = halton - h
        norm = np.linalg.norm(diff)
        self.assertAlmostEqual(norm, 0, 7)
        halton = dr.get_halton_draws(sample_size=1, number_of_draws=10, base=3, skip=10)
        h = [
            0.7037037,
            0.14814815,
            0.48148148,
            0.81481481,
            0.25925926,
            0.59259259,
            0.92592593,
            0.07407407,
            0.40740741,
            0.74074074,
        ]
        diff = halton - h
        norm = np.linalg.norm(diff)
        self.assertAlmostEqual(norm, 0, 7)

    def test_halton_shuffled(self):
        halton = dr.get_halton_draws(
            sample_size=1, number_of_draws=100, base=5, shuffled=True
        )
        self.assertEqual(halton.size, 100)

    def test_halton_symmetric(self):
        halton = dr.get_halton_draws(
            sample_size=1, number_of_draws=100, base=5, symmetric=True
        )
        self.assertEqual(halton.size, 100)

    def test_halton_exception(self):
        with self.assertRaises(excep.BiogemeError):
            halton = dr.get_halton_draws(sample_size=1, number_of_draws=-100, base=3)
        with self.assertRaises(excep.BiogemeError):
            halton = dr.get_halton_draws(sample_size=-1, number_of_draws=100, base=3)

    def test_antithetic(self):
        draws = dr.get_antithetic(
            dr.get_halton_draws, sample_size=1, number_of_draws=10
        )
        d = [0.5, 0.25, 0.75, 0.125, 0.625, 0.5, 0.75, 0.25, 0.875, 0.375]
        self.assertListEqual(d, draws[0].tolist())

    def test_normal(self):
        draws = dr.get_normal_wichura_draws(sample_size=3, number_of_draws=1000000)
        mean = np.linalg.norm(np.average(draws, axis=1))
        self.assertAlmostEqual(mean, 0, 2)
        draws = dr.get_normal_wichura_draws(
            sample_size=3, number_of_draws=1000000, antithetic=True
        )
        mean = np.linalg.norm(np.average(draws, axis=1))
        self.assertAlmostEqual(mean, 0, 5)

    def test_normal_exception(self):
        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_normal_wichura_draws(sample_size=3, number_of_draws=-10)

        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_normal_wichura_draws(
                sample_size=3, number_of_draws=9, antithetic=True
            )

        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_normal_wichura_draws(sample_size=-3, number_of_draws=10)

        uniform_numbers = np.random.uniform(size=2)
        with self.assertRaises(excep.BiogemeError):
            draws = dr.get_normal_wichura_draws(
                sample_size=3, number_of_draws=10, uniform_numbers=uniform_numbers
            )


if __name__ == '__main__':
    unittest.main()



"""Arithmetic expressions accepted by Biogeme: power constant

Michel Bierlaire
Sat Jun 28 2025, 12:20:48
"""

from __future__ import annotations

import logging

import jax
import jax.numpy as jnp
import numpy as np
from biogeme.exceptions import BiogemeError
from biogeme.floating_point import JAX_FLOAT

from .base_expressions import ExpressionOrNumeric
from .jax_utils import JaxFunctionType
from .unary_expressions import UnaryOperator

logger = logging.getLogger(__name__)


class PowerConstant(UnaryOperator):
    """
    Raise the argument to a constant power.
    """

    def __init__(self, child: ExpressionOrNumeric, exponent: float):
        """Constructor

        :param child: first arithmetic expression
        :type child: biogeme.expressions.Expression
        """
        super().__init__(child)
        self.exponent: float = exponent
        epsilon = np.finfo(float).eps
        if abs(exponent - round(exponent)) < epsilon:
            self.integer_exponent = int(round(exponent))
        else:
            self.integer_exponent = None

    def deep_flat_copy(self) -> PowerConstant:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_child = self.child.deep_flat_copy()
        return type(self)(child=copy_child, exponent=self.exponent)

    def __str__(self) -> str:
        return f'{self.child}**{self.exponent}'

    def __repr__(self) -> str:
        return f'PowerConstant({repr(self.child)}, {self.exponent})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        v = self.child.get_value()
        if v == 0:
            return 0.0
        if v > 0:
            return v**self.exponent
        if self.integer_exponent is not None:
            return v**self.integer_exponent
        if v < 0:
            error_msg = f'Cannot calculate {v}**{self.exponent}'
            raise BiogemeError(error_msg)

        return 0 if v == 0 else np.log(v)

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        child_jax = self.child.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        if numerically_safe:

            def the_jax_function(
                parameters: jnp.ndarray,
                one_row: jnp.ndarray,
                the_draws: jnp.ndarray,
                the_random_variables: jnp.ndarray,
            ) -> jnp.ndarray:
                child_value = child_jax(
                    parameters, one_row, the_draws, the_random_variables
                )
                epsilon = jnp.finfo(JAX_FLOAT).eps

                if self.integer_exponent is not None:
                    abs_exponent = jnp.abs(self.integer_exponent)
                    safe_value = jnp.sqrt(child_value**2 + epsilon)
                    powered = safe_value**abs_exponent
                    signed = jnp.where(
                        child_value < 0, (-1) ** self.integer_exponent, 1.0
                    )
                    result = jnp.where(self.exponent < 0, 1.0 / powered, powered)

                    near_zero = jnp.logical_and(
                        child_value >= -epsilon, child_value <= epsilon
                    )

                    def zero_case(_):
                        return jnp.array(0.0, dtype=JAX_FLOAT)

                    def nonzero_case(_):
                        return result * signed

                    return jax.lax.cond(
                        near_zero if self.integer_exponent > 0 else child_value == 0.0,
                        zero_case,
                        nonzero_case,
                        operand=None,
                    )
                else:

                    def nan_branch(_):
                        return jnp.nan

                    def safe_branch(_):
                        return jnp.exp(
                            self.exponent
                            * jnp.log(jnp.clip(child_value, a_min=epsilon))
                        )

                    return jax.lax.cond(
                        child_value == 0.0,
                        lambda _: jnp.array(0.0, dtype=JAX_FLOAT),
                        lambda _: jax.lax.cond(
                            child_value < 0.0, nan_branch, safe_branch, operand=None
                        ),
                        operand=None,
                    )

            return the_jax_function

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            child_value = child_jax(
                parameters, one_row, the_draws, the_random_variables
            )
            return jnp.power(child_value, self.exponent)

        return the_jax_function



"""

Estimation and simulation of a nested logit model
=================================================

 We estimate a nested logit model and we perform simulation using the
 estimated model.

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:05:16 2023

"""
from biogeme import models
import biogeme.biogeme as bio
from optima_data import database
from scenarios import scenario

# %%
# Obtain the specification for the default scenario.
# The definition of the scenarios is available in :ref:`scenarios`.
V, nests, Choice, _ = scenario()

# %%
# The choice model is a nested logit, with availability conditions
# For estimation, we need the log of the probability.
logprob = models.lognested(V, None, nests, Choice)

# %%
# Create the Biogeme object for estimation.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b02estimation'

# %%
# Estimate the parameters. Perform bootstrapping.
the_biogeme.bootstrap_samples = 100
results = the_biogeme.estimate(run_bootstrap=True)

# %%
# Get the results in a pandas table
pandas_results = results.getEstimatedParameters()
pandas_results


# %%
# Simulation
simulated_choices = logprob.getValue_c(betas=results.getBetaValues(), database=database)
simulated_choices

# %%
loglikelihood = logprob.getValue_c(
    betas=results.getBetaValues(),
    database=database,
    aggregation=True,
)
print(f'Final log likelihood:     {results.data.logLike}')
print(f'Simulated log likelihood: {loglikelihood}')



"""
Test executing all examples and verifying that they run properly

:author: Michel Bierlaire
:date: Sat Jul  8 17:22:54 2023

"""

import os
import subprocess
import shutil
import unittest
import tempfile
from biogeme.parameters import biogeme_parameters

ROOT_DIR = (
    '/Users/bierlair/ToBackupOnGoogleDrive/FilesOnGoogleDrive/GitHub/biogeme/examples'
)


exclude = ['replace.py']

examples_dirs = (
    'swissmetro',
    'swissmetro_panel',
    'wtp_space',
    'assisted',
    'indicators',
    'latent',
    'latentbis',
    'montecarlo',
    'sampling',
)

extensions_to_clean = ['html', 'pickle', 'iter', 'log', 'pareto']


def clean_directory(directory):
    file_list = os.listdir(directory)
    for file in file_list:
        for ext in extensions_to_clean:
            if file.endswith(ext):
                file_path = os.path.join(directory, file)
                print(f'Remove {file_path}')
                os.remove(file_path)


class ScriptExecutionTests(unittest.TestCase):
    def setUp(self):

        self.filename = tempfile.mktemp()
        biogeme_parameters.set_value(
            name='number_of_draws', value=4, section='MonteCarlo'
        )
        biogeme_parameters.set_value(
            name='largest_neighborhood', value=2, section='AssistedSpecification'
        )
        biogeme_parameters.set_value(
            name='maximum_attempts', value=2, section='AssistedSpecification'
        )
        biogeme_parameters.set_value(
            name='number_of_neighbors', value=2, section='AssistedSpecification'
        )
        biogeme_parameters.set_value(
            name='max_iterations', value=2, section='SimpleBounds'
        )
        biogeme_parameters.set_value(
            name='bootstrap_samples', value=2, section='Estimation'
        )

        biogeme_parameters.set_value(
            name='second_derivatives', value=0, section='SimpleBounds'
        )

        biogeme_parameters.dump_file(file_name=self.filename)

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_script_execution(self):

        # Iterate over all directories
        for directory in examples_dirs:
            # Clean the directory
            script_dir = f'{ROOT_DIR}/{directory}'
            clean_directory(script_dir)
            toml_file = f'{script_dir}/biogeme.toml'
            orig_file_exists = False
            # Replace the file biogeme.toml, if present, with another file
            if os.path.isfile(toml_file):
                orig_file_exists = True
                backup_file = f'{script_dir}/biogeme.toml.backup'
                shutil.move(toml_file, backup_file)

            shutil.copy2(self.filename, toml_file)

            # Iterate over all .py in the directory
            os.chdir(script_dir)
            for file_name in sorted(os.listdir(script_dir)):
                if file_name in exclude:
                    continue
                file_path = os.path.join(script_dir, file_name)
                # Skip directories and non-Python .py
                if not os.path.isfile(file_path) or not file_name.endswith('.py'):
                    continue

                print(f'Run {file_path}...')

                # Execute the script using subprocess
                try:
                    subprocess.check_output(
                        ['python', file_path], stderr=subprocess.STDOUT
                    )
                except subprocess.CalledProcessError as e:
                    self.fail(
                        f"Script {file_name} failed with error: {e.output.decode()}"
                    )
            if orig_file_exists:
                shutil.move(backup_file, toml_file)
            else:
                os.remove(toml_file)


if __name__ == '__main__':
    unittest.main()



"""File mengyi_data.py

:author: Michel Bierlaire, EPFL
:date: 

First attempt to estimate a MDCEV model

"""

import os

import pandas as pd
import biogeme.database as db
from biogeme.expressions import Variable


def read_data() -> db.Database:
    """Read the data from file"""
    # Get the directory of the current file
    module_dir = os.path.dirname(__file__)

    # Construct the path to the data file
    data_file_path = os.path.join(module_dir, 'data', 'mdcev.csv')

    # %%
    # Read the data.
    df = pd.read_csv(data_file_path, sep='\t')
    database = db.Database('mdcev', df)
    return database


PersonID = Variable('PersonID')
weight = Variable('weight')
hhsize = Variable('hhsize')
childnum = Variable('childnum')
faminc = Variable('faminc')
faminc25K = Variable('faminc25K')
income = Variable('income')
employed = Variable('employed')
fulltime = Variable('fulltime')
spousepr = Variable('spousepr')
spousemp = Variable('spousemp')
male = Variable('male')
married = Variable('married')
age = Variable('age')
age2 = Variable('age2')
age15_40 = Variable('age15_40')
age41_60 = Variable('age41_60')
age61_85 = Variable('age61_85')
bachigher = Variable('bachigher')
white = Variable('white')
metro = Variable('metro')
diaryday = Variable('diaryday')
Sunday = Variable('Sunday')
holiday = Variable('holiday')
weekearn = Variable('weekearn')
weekwordur = Variable('weekwordur')
hhchild = Variable('hhchild')
ohhchild = Variable('ohhchild')
t1 = Variable('t1')
t2 = Variable('t2')
t3 = Variable('t3')
t4 = Variable('t4')
number_chosen = Variable('number_chosen')



"""Tests for Expression

Michel Bierlaire
Tue Mar 25 17:43:42 2025
"""

import unittest

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, list_of_all_betas_in_expression
from biogeme.expressions.rename_variables import rename_all_variables


class DummyExpression(Expression):
    def __init__(self, label="dummy"):
        super().__init__()
        self.label = label

    def __str__(self):
        return self.label

    def recursive_construct_jax_function(self):
        raise NotImplementedError

    def get_signature(self):
        return [f"<Dummy>{{{id(self)}}}({len(self.children)})".encode()]


class TestExpression(unittest.TestCase):
    def setUp(self):
        self.expr1 = DummyExpression("A")
        self.expr2 = DummyExpression("B")

    def test_addition(self):
        result = self.expr1 + self.expr2
        self.assertEqual(result.get_class_name(), "Plus")

    def test_subtraction(self):
        result = self.expr1 - self.expr2
        self.assertEqual(result.get_class_name(), "Minus")

    def test_multiplication(self):
        result = self.expr1 * self.expr2
        self.assertEqual(result.get_class_name(), "Times")

    def test_division(self):
        result = self.expr1 / self.expr2
        self.assertEqual(result.get_class_name(), "Divide")

    def test_unary_minus(self):
        result = -self.expr1
        self.assertEqual(result.get_class_name(), "UnaryMinus")

    def test_power_expression(self):
        result = self.expr1**self.expr2
        self.assertEqual(result.get_class_name(), "Power")

    def test_power_constant(self):
        result = self.expr1**2
        self.assertEqual(result.get_class_name(), "PowerConstant")

    def test_logical_and(self):
        result = self.expr1 & self.expr2
        self.assertEqual(result.get_class_name(), "And")

    def test_logical_or(self):
        result = self.expr1 | self.expr2
        self.assertEqual(result.get_class_name(), "Or")

    def test_comparisons(self):
        self.assertEqual((self.expr1 == self.expr2).get_class_name(), "Equal")
        self.assertEqual((self.expr1 != self.expr2).get_class_name(), "NotEqual")
        self.assertEqual((self.expr1 < self.expr2).get_class_name(), "Less")
        self.assertEqual((self.expr1 <= self.expr2).get_class_name(), "LessOrEqual")
        self.assertEqual((self.expr1 > self.expr2).get_class_name(), "Greater")
        self.assertEqual((self.expr1 >= self.expr2).get_class_name(), "GreaterOrEqual")

    def test_set_and_dict_of_elementary_expression(self):
        expr = DummyExpression()
        the_betas = list_of_all_betas_in_expression(expr)
        self.assertFalse(the_betas)

    def test_get_elementary_expression_none(self):
        self.assertIsNone(self.expr1.get_elementary_expression("nonexistent"))

    def test_rename_elementary_does_not_fail(self):
        try:
            _ = rename_all_variables(
                expr=self.expr1, old_name='x', new_name='pre_x_suf'
            )
            _ = rename_all_variables(
                expr=self.expr1,
                old_name='y',
                new_name='pre_y_suf',
            )
        except Exception as e:
            self.fail(f"rename_elementary raised an exception unexpectedly: {e}")

    def test_boolean_context_raises(self):
        with self.assertRaises(BiogemeError):
            bool(self.expr1)


if __name__ == '__main__':
    unittest.main()



"""Arithmetic expressions accepted by Biogeme: Monte-Carlo integration

Michel Bierlaire
Thu Apr 24 2025, 18:47:01
"""

from __future__ import annotations

import logging

import jax.numpy as jnp
from jax import vmap

from .base_expressions import ExpressionOrNumeric
from .jax_utils import JaxFunctionType
from .unary_expressions import UnaryOperator

logger = logging.getLogger(__name__)


class MonteCarlo(UnaryOperator):
    """
    Monte Carlo integration
    """

    def __init__(self, child: ExpressionOrNumeric):
        """Constructor

        :param child: arithmetic expression
        :type child: biogeme.expressions.Expression
        """
        super().__init__(child)

    def deep_flat_copy(self) -> MonteCarlo:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_child = self.child.deep_flat_copy()
        return type(self)(child=copy_child)

    def __str__(self) -> str:
        return f'MonteCarlo({self.child})'

    def __repr__(self) -> str:
        return f'MonteCarlo({repr(self.child)})'

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        child_jax = self.child.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        vectorized_function = vmap(
            lambda parameters, row, draws, random_variables: child_jax(
                parameters, row, draws, random_variables
            ),
            in_axes=(None, None, 0, None),
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            value_per_draw = vectorized_function(
                parameters, one_row, the_draws, the_random_variables
            )
            mean = jnp.mean(value_per_draw, axis=-1)
            return mean

        return the_jax_function



"""

Logit
=====

Estimation of a logit model using sampling of alternatives.

:author: Michel Bierlaire
:date: Wed Nov  1 17:39:47 2023
"""

import pandas as pd
from biogeme.sampling_of_alternatives import (
    SamplingContext,
    ChoiceSetsGeneration,
    GenerateModel,
    generate_segment_size,
)
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from compare import compare
from specification import V, combined_variables
from alternatives import (
    alternatives,
    ID_COLUMN,
    partitions,
)

# %%
logger = blog.get_screen_logger(level=blog.INFO)

# %%
# The data file contains several columns associated with synthetic
# choices. Here we arbitrarily select `logit_4`.
CHOICE_COLUMN = 'logit_4'

# %%
SAMPLE_SIZE = 10
PARTITION = 'asian'
MODEL_NAME = f'logit_{PARTITION}_{SAMPLE_SIZE}_alt'
FILE_NAME = f'{MODEL_NAME}.dat'
OBS_FILE = 'obs_choice.dat'

# %%
the_partition = partitions.get(PARTITION)
if the_partition is None:
    raise ValueError(f'Unknown partition: {PARTITION}')

# %%
segment_sizes = generate_segment_size(SAMPLE_SIZE, the_partition.number_of_segments())

# %%
observations = pd.read_csv(OBS_FILE)

# %%
context = SamplingContext(
    the_partition=the_partition,
    sample_sizes=segment_sizes,
    individuals=observations,
    choice_column=CHOICE_COLUMN,
    alternatives=alternatives,
    id_column=ID_COLUMN,
    biogeme_file_name=FILE_NAME,
    utility_function=V,
    combined_variables=combined_variables,
)

# %%
logger.info(context.reporting())

# %%
the_data_generation = ChoiceSetsGeneration(context=context)

# %%
the_model_generation = GenerateModel(context=context)

# %%
biogeme_database = the_data_generation.sample_and_merge(recycle=False)

# %%
logprob = the_model_generation.get_logit()

# %%
the_biogeme = bio.BIOGEME(biogeme_database, logprob)
the_biogeme.modelName = MODEL_NAME

# %%
# Calculate the null log likelihood for reporting.
the_biogeme.calculateNullLoglikelihood({i: 1 for i in range(SAMPLE_SIZE)})

# %%
# Estimate the parameters
results = the_biogeme.estimate(recycle=False)

# %%
print(results.short_summary())

# %%
estimated_parameters = results.getEstimatedParameters()
estimated_parameters

# %%
df, msg = compare(estimated_parameters)

# %%
print(df)

# %%
print(msg)



"""Tests for elementary expressions

Michel Bierlaire
Tue Mar 25 17:43:42 2025
"""

import unittest
import warnings

import jax.numpy as jnp

from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Draws,
    RandomVariable,
    TypeOfElementaryExpression,
    Variable,
    bioDraws,
)
from biogeme.expressions.elementary_expressions import Elementary


class TestElementaryExpressions(unittest.TestCase):
    def test_elementary_repr_and_str(self):
        e = Elementary("x")
        e.elementary_index = 0
        self.assertEqual(str(e), "x")
        self.assertIn("Elementary", repr(e))

    def test_elementary_dict_type_match(self):
        e = Elementary("x")
        e.expression_type = TypeOfElementaryExpression.VARIABLE
        result = e.dict_of_elementary_expression(TypeOfElementaryExpression.VARIABLE)
        self.assertIn("x", result)

    def test_elementary_dict_type_mismatch(self):
        e = Elementary("x")
        e.expression_type = TypeOfElementaryExpression.DRAWS
        result = e.dict_of_elementary_expression(TypeOfElementaryExpression.VARIABLE)
        self.assertEqual(result, {})

    def test_variable_jax_function(self):
        v = Variable("income")
        v.specific_id = 2
        jax_fn = v.recursive_construct_jax_function(numerically_safe=False)
        one_row = jnp.array([10.0, 20.0, 30.0])
        result = jax_fn(None, one_row, None, None)
        self.assertEqual(result, 30.0)

    def test_variable_missing_id(self):
        v = Variable("age")
        with self.assertRaises(BiogemeError):
            _ = v.safe_variable_id

    def test_biodraws_jax_function(self):
        d = Draws("eps", "NORMAL")
        d.specific_id = 1
        jax_fn = d.recursive_construct_jax_function(numerically_safe=False)
        draws = jnp.array([[0.0, 0.5, 1.0]])
        result = jax_fn(None, None, draws, None)
        self.assertAlmostEqual(result, 0.5)

    def test_biodraws_missing_id(self):
        d = Draws("eps", "NORMAL")
        with self.assertRaises(BiogemeError):
            _ = d.safe_draw_id

    def test_random_variable_id_access(self):
        rv = RandomVariable("z")
        rv.rv_id = 0
        self.assertEqual(rv.rv_id, 0)

    def test_random_variable_missing_id(self):
        rv = RandomVariable("z")
        with self.assertRaises(BiogemeError):
            _ = rv.safe_rv_id

    def test_biodraws_deprecation_warning(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")  # ensure all warnings are caught

            _ = bioDraws("eps", "NORMAL")

            self.assertTrue(
                any(issubclass(warning.category, DeprecationWarning) for warning in w)
            )
            self.assertTrue(
                any("deprecated" in str(warning.message).lower() for warning in w)
            )



"""

Choice model with a latent variable: sequential estimation (Monte-Carlo)
========================================================================

Mixture of logit, with Monte-Carlo integration
Measurement equation for the indicators.
Sequential estimation.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 18:00:05 2023

"""

import sys
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
import biogeme.exceptions as excep
from biogeme import models
from biogeme.results import bioResults
from biogeme.expressions import (
    Beta,
    bioDraws,
    MonteCarlo,
    exp,
    log,
)

from read_or_estimate import read_or_estimate

from optima import (
    database,
    age_65_more,
    moreThanOneCar,
    moreThanOneBike,
    individualHouse,
    male,
    haveChildren,
    haveGA,
    highEducation,
    WaitingTimePT,
    Choice,
    TimePT_scaled,
    TimeCar_scaled,
    MarginalCostPT_scaled,
    CostCarCHF_scaled,
    distance_km_scaled,
    PurpHWH,
    PurpOther,
    ScaledIncome,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b04latent_choice_seq_mc.py')

# %%
# Read the estimates from the structural equation estimation.
MODELNAME = 'b02one_latent_ordered'
try:
    struct_results = bioResults(pickleFile=f'saved_results/{MODELNAME}.pickle')
except excep.BiogemeError:
    print(
        f'Run first the script {MODELNAME}.py in order to generate the '
        f'file {MODELNAME}.pickle, and move it to the directory saved_results'
    )
    sys.exit()
struct_betas = struct_results.getBetaValues()

# %%
# Coefficients.
coef_intercept = struct_betas['coef_intercept']
coef_age_65_more = struct_betas['coef_age_65_more']
coef_haveGA = struct_betas['coef_haveGA']
coef_moreThanOneCar = struct_betas['coef_moreThanOneCar']
coef_moreThanOneBike = struct_betas['coef_moreThanOneBike']
coef_individualHouse = struct_betas['coef_individualHouse']
coef_male = struct_betas['coef_male']
coef_haveChildren = struct_betas['coef_haveChildren']
coef_highEducation = struct_betas['coef_highEducation']

# %%
# Latent variable: structural equation.

# %%
# Define a random parameter, normally distributed, designed to be used
# for numerical integration

sigma_s = Beta('sigma_s', 1, None, None, 0)
error_component = sigma_s * bioDraws('EC', 'NORMAL_MLHS')

# %%
# Piecewise linear specification for income.
thresholds = [None, 4, 6, 8, 10, None]
formula_income = models.piecewiseFormula(
    variable=ScaledIncome,
    thresholds=thresholds,
    betas=[
        struct_betas['beta_ScaledIncome_minus_inf_4'],
        struct_betas['beta_ScaledIncome_4_6'],
        struct_betas['beta_ScaledIncome_6_8'],
        struct_betas['beta_ScaledIncome_8_10'],
        struct_betas['beta_ScaledIncome_10_inf'],
    ],
)


CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formula_income
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
    + error_component
)

# %%
# Choice model.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0, None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', 0, None, None, 0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', 0, None, None, 0)
BETA_DIST = Beta('BETA_DIST', 0, None, None, 0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF', 0, None, 0, 0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', 0, None, 0, 0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME', 0, None, None, 0)

# %%
# The coefficient of the latent variable should be initialized to
# something different from zero. If not, the algorithm may be trapped
# in a local optimum, and never change the value.
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL', -0.01, None, None, 0)
BETA_TIME_CAR_CL = Beta('BETA_TIME_CAR_CL', -0.01, None, None, 0)

# %%
# Definition of utility functions:.
BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# %%
# Associate utility functions with the numbering of alternatives.
V = {0: V0, 1: V1, 2: V2}

# %%
# Conditional on omega, we have a logit model (called the kernel).
condprob = models.logit(V, None, Choice)

# %%
# We integrate over omega using numerical integration
loglike = log(MonteCarlo(condprob))

#
# Create the Biogeme object. As the objective is to illustrate the
# syntax, we calculate the Monte-Carlo approximation with a small
# number of draws. To achieve that, we provide a parameter file
# different from the default one: `few_draws.toml<few_draws.toml>`_
the_biogeme = bio.BIOGEME(database, loglike, parameter_file='few_draws.toml')
the_biogeme.modelName = 'b04latent_choice_seq_mc'

# %%
# If estimation results are saved on file, we read them to speed up the process.
# If not, we estimate the parameters.
results = read_or_estimate(the_biogeme=the_biogeme, directory='saved_results')

# %%
print(f'Estimated betas: {len(results.data.betaValues)}')
print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')
results.writeLaTeX()
print(f'LaTeX file: {results.data.latexFileName}')

# %%
results.getEstimatedParameters()



""".. _plot_b21multiple_models:

Assisted specification
======================

 Example of the estimation of several versions of the model using
 assisted specification algorithm. The catalog of specifications is
 defined in :ref:`plot_b21multiple_models_spec` . All specifications
 are estimated. Have a look at :ref:`plot_b22multiple_models` for an
 example where the number of specifications is too high to be
 enumerated.

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 16:58:49 2023

"""
import biogeme.biogeme_logging as blog
from biogeme.results import compile_estimation_results
from biogeme.multiobjectives import loglikelihood_dimension
from biogeme.assisted import AssistedSpecification
from plot_b21multiple_models_spec import the_biogeme, PARETO_FILE_NAME

logger = blog.get_screen_logger(blog.INFO)
logger.info('Example b21multipleModels')

# %%
# Creation of the object capturing the assisted specification algorithm.
# Its constructor takes three arguments:
#
#    - the biogeme object containing the specifications and the
#      database,
#    - an object defining the objectives to minimize. Here, we use the
#      opposite of the log lieklihood and the number of estimated
#      parameters.
#    - the name of the file where the estimated are saved, and
#      organized into a Pareto set.
assisted_specification = AssistedSpecification(
    biogeme_object=the_biogeme,
    multi_objectives=loglikelihood_dimension,
    pareto_file_name=PARETO_FILE_NAME,
)

# %%
# The algorithm is run.
non_dominated_models = assisted_specification.run()

# %%
summary, description = compile_estimation_results(
    non_dominated_models, use_short_names=True
)
print(summary)

# %%
# Explanation of the short names of the model.
for k, v in description.items():
    if k != v:
        print(f'{k}: {v} AIC={summary.at["Akaike Information Criterion", k]}')



"""
Simple estimation test from the Ben-Akiva and Lerman book

:author: Michel Bierlaire
:date: Wed Apr 29 18:43:34 2020

"""

import shutil
import tempfile
import unittest
from os import path

import pandas as pd

import biogeme.biogeme as bio
import biogeme.database as db
from biogeme import models
from biogeme.expressions import Beta, Variable

GROUP = Variable('GROUP')
SURVEY = Variable('SURVEY')
SP = Variable('SP')
ID = Variable('ID')
PURPOSE = Variable('PURPOSE')
FIRST = Variable('FIRST')
TICKET = Variable('TICKET')
WHO = Variable('WHO')
LUGGAGE = Variable('LUGGAGE')
AGE = Variable('AGE')
MALE = Variable('MALE')
INCOME = Variable('INCOME')
GA = Variable('GA')
ORIGIN = Variable('ORIGIN')
DEST = Variable('DEST')
TRAIN_AV = Variable('TRAIN_AV')
CAR_AV = Variable('CAR_AV')
SM_AV = Variable('SM_AV')
TRAIN_TT = Variable('TRAIN_TT')
TRAIN_CO = Variable('TRAIN_CO')
TRAIN_HE = Variable('TRAIN_HE')
SM_TT = Variable('SM_TT')
SM_CO = Variable('SM_CO')
SM_HE = Variable('SM_HE')
SM_SEATS = Variable('SM_SEATS')
CAR_TT = Variable('CAR_TT')
CAR_CO = Variable('CAR_CO')
CHOICE = Variable('CHOICE')


class test_biogeme(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.toml_file = path.join(self.test_dir, 'biogeme.toml')

        data = {
            'ID': pd.Series([i + 1 for i in range(21)]),
            'auto_time': pd.Series(
                [
                    52.9,
                    4.1,
                    4.1,
                    56.2,
                    51.8,
                    0.2,
                    27.6,
                    89.9,
                    41.5,
                    95.0,
                    99.1,
                    18.5,
                    82.0,
                    8.6,
                    22.5,
                    51.4,
                    81.0,
                    51.0,
                    62.2,
                    95.1,
                    41.6,
                ]
            ),
            'transit_time': pd.Series(
                [
                    4.4,
                    28.5,
                    86.9,
                    31.6,
                    20.2,
                    91.2,
                    79.7,
                    2.2,
                    24.5,
                    43.5,
                    8.4,
                    84.0,
                    38.0,
                    1.6,
                    74.1,
                    83.8,
                    19.2,
                    85.0,
                    90.1,
                    22.2,
                    91.5,
                ]
            ),
            'Choice': pd.Series(
                [1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0]
            ),
        }
        dataframe = pd.DataFrame(data)

        self.database = db.Database('akiva', dataframe)

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def testEstimation(self):
        AutoTime = Variable('auto_time')
        TransitTime = Variable('transit_time')
        Choice = Variable('Choice')

        ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
        B_TIME = Beta('B_TIME', 0, None, None, 0)

        V = {0: ASC_CAR + B_TIME * AutoTime, 1: B_TIME * TransitTime}

        av = {0: 1, 1: 1}

        logprob = models.loglogit(V, av, Choice)

        biogeme = bio.BIOGEME(
            self.database,
            logprob,
            parameters=self.toml_file,
            generate_html=False,
            generate_yaml=False,
            save_iterations=False,
            optimization_algorithm='scipy',
        )
        biogeme.model_name = 'test'
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -6.166042, 2)


if __name__ == '__main__':
    unittest.main()



"""Defines controllers for multiple expressions.

:author: Michel Bierlaire
:date: Sun Jul 16 15:23:46 2023

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Iterable, Callable

from biogeme.expressions import SELECTION_SEPARATOR
from biogeme.exceptions import BiogemeError
from .configuration import (
    Configuration,
)

if TYPE_CHECKING:
    from biogeme.catalog import Catalog

ControllerOperator = Callable[
    [Configuration, int],
    tuple[Configuration, int],
]

logger = logging.getLogger(__name__)


class Controller:
    """Class controlling the specification of the Catalogs"""

    def __init__(self, controller_name: str, specification_names: Iterable[str]):
        """Constructor

        :param controller_name: name of the controller
        :type controller_name: str

        :param specification_names: list or tuple of the names of the
            specification controlled by the controller
        :type specification_names: list(str) or tuple(str)
        """
        self.controller_name: str = controller_name
        self.specification_names: tuple[str, ...] = tuple(specification_names)
        self.current_index: int = 0
        self.dict_of_index: dict[str, int] = {
            name: index for index, name in enumerate(self.specification_names)
        }
        self.controlled_catalogs: list[Catalog] = []

    def all_configurations(self) -> set[str]:
        """Return the code of all configurations

        :return: set of codes
        :rtype: set(str)
        """
        return {
            f"{self.controller_name}{SELECTION_SEPARATOR}{specification}"
            for specification in self.specification_names
        }

    def __eq__(self, other: Controller) -> bool:
        return self.controller_name == other.controller_name

    def __lt__(self, other: Controller) -> bool:
        return self.controller_name < other.controller_name

    def __hash__(self) -> int:
        return hash(self.controller_name)

    def __repr__(self) -> str:
        return repr(self.controller_name)

    def __str__(self) -> str:
        return f"{self.controller_name}: {self.specification_names}"

    def controller_size(self) -> int:
        """Number of specifications managed by this controller"""
        return len(self.specification_names)

    def current_name(self) -> str:
        """Name of the currently selected expression"""
        return self.specification_names[self.current_index]

    def set_name(self, name: str) -> None:
        """Set the index of the controller based on the name of the specification

        :param name: name of the specification
        :type name: str
        """
        the_index = self.dict_of_index.get(name)
        if the_index is None:
            error_msg = (
                f"{name}: unknown specification for controller {self.controller_name}"
            )
            raise BiogemeError(error_msg)
        self.set_index(the_index)

    def set_index(self, index: int) -> None:
        """Set the index of the controller, and update the controlled catalogs

        :param index: value of the index
        :type index: int

        :raises BiogemeError: if index is out of range
        """
        if index < 0 or index >= self.controller_size():
            error_msg = (
                f"Wrong index {index} for controller {self.controller_name}. "
                f"Must be in [0, {self.controller_size()}]"
            )
            raise BiogemeError(error_msg)

        self.current_index = index
        for catalog in self.controlled_catalogs:
            catalog.current_index = index

    def reset_selection(self) -> None:
        """Select the first specification"""
        self.set_index(0)

    def modify_controller(self, step: int, circular: bool) -> int:
        """Modify the specification of the controller

        :param step: increment of the modifications. Can be negative.
        :type step: int

        :param circular: If True, the modification is always made. If
            the selection needs to move past the last one, it comes
            back to the first one. For instance, if the catalog is
            currently at its last value, and the step is 1, it is set
            to its first value. If circular is False, and the
            selection needs to move past the last one, the selection
            is set to the last one. It works symmetrically if the step
            is negative
        :type circular: bool

        :return: number of actual modifications
        :rtype: int

        """
        the_size = self.controller_size()
        new_index = self.current_index + step
        if circular:
            self.set_index(new_index % the_size)
            return step

        if new_index < 0:
            total_modif = self.current_index
            self.set_index(0)
            return total_modif

        if new_index >= the_size:
            total_modif = the_size - 1 - self.current_index
            self.set_index(the_size - 1)
            return total_modif

        self.set_index(new_index)
        return step



"""
Run a model estimation using a specified optimization algorithm.

This module defines the interface for executing a Biogeme-compatible
optimization routine and collecting its results.

Michel Bierlaire
Sun Mar 30 16:07:55 2025
"""

import logging
from datetime import datetime
from typing import Any, NamedTuple

import numpy as np
from biogeme_optimization.function import FunctionToMinimize

from biogeme.calculator import (
    CallableExpression,
    CompiledFormulaEvaluator,
    function_from_compiled_formula,
)
from biogeme.default_parameters import ParameterValue
from biogeme.likelihood.negative_likelihood import NegativeLikelihood
from biogeme.optimization import OptimizationAlgorithm

logger = logging.getLogger(__name__)


class AlgorithmResults(NamedTuple):
    """
    Container for the results returned by an optimization algorithm.

    :param solution: Optimal values of the parameters as a NumPy array.
    :param optimization_messages: Dictionary with diagnostic messages from the optimizer.
    :param convergence: Boolean indicating whether the optimization terminated successfully.
    """

    solution: np.ndarray
    optimization_messages: dict[str, Any]
    convergence: bool


def optimization(
    the_algorithm: OptimizationAlgorithm,
    the_function: FunctionToMinimize,
    starting_values: np.ndarray,
    bounds: list[tuple[float, float]],
    variable_names: list[str],
    parameters: dict[str, Any],
) -> AlgorithmResults:
    """
    Run an optimization algorithm to estimate model parameters.

    :param the_algorithm: Optimization algorithm conforming to the Biogeme interface.
    :param the_function: Function to minimize, providing function value and derivatives.
    :param starting_values: Initial guess for the optimization variables.
    :param bounds: List of (lower, upper) bounds for each parameter.
    :param variable_names: Names of the variables (used for reporting or algorithm diagnostics).
    :param parameters: Dictionary of additional parameters passed to the optimization algorithm.

    :return: A tuple with:
        - x_star: the optimal solution (array of estimated parameters),
        - optimization_messages: a dictionary with diagnostic messages and timing,
        - convergence: a boolean indicating whether the optimization converged successfully.
    """
    the_function.set_variables(starting_values)
    start_time = datetime.now()
    output = the_algorithm(
        fct=the_function,
        init_betas=starting_values,
        bounds=bounds,
        variable_names=variable_names,
        parameters=parameters,
    )
    x_star, optimization_messages, convergence = output
    optimization_messages["Optimization time"] = datetime.now() - start_time
    return AlgorithmResults(
        solution=x_star,
        optimization_messages=optimization_messages,
        convergence=convergence,
    )


def model_estimation(
    the_algorithm: OptimizationAlgorithm,
    function_evaluator: CompiledFormulaEvaluator,
    parameters: dict[str, ParameterValue],
    some_starting_values: dict[str, float],
    save_iterations_filename: str | None,
) -> AlgorithmResults:
    """
    Estimate a model using the specified optimization algorithm and modeling elements.

    This function prepares the model log-likelihood function and its derivatives based
    on the provided modeling elements and starting values. It constructs the objective
    function and delegates the actual optimization to the `optimization` routine.

    :param the_algorithm: The optimization algorithm to use.
    :param function_evaluator: Object with the compiled information to evaluate the function.
    :param parameters: Dictionary of configuration parameters for the estimation.
    :param some_starting_values: Initial values for a subset or all of the model's free parameters.
    :param save_iterations_filename: If not None, the name of the file where to save the best iterations.

    :return: A tuple containing:
        - the optimal parameter values as a NumPy array,
        - a dictionary of optimization diagnostic messages,
        - a boolean indicating whether the optimization converged successfully.
    """
    starting_values = function_evaluator.model_elements.expressions_registry.complete_dict_of_free_beta_values(
        the_betas=some_starting_values
    )

    the_function: CallableExpression = function_from_compiled_formula(
        the_compiled_function=function_evaluator,
        the_betas=starting_values,
    )

    the_function_to_minimize = NegativeLikelihood(
        dimension=function_evaluator.model_elements.expressions_registry.number_of_free_betas,
        loglikelihood=the_function,
        parameters=parameters,
    )

    the_function_to_minimize.set_variables(np.array(list(starting_values.values())))
    if save_iterations_filename is not None:
        the_function_to_minimize.save_iterations(
            filename_for_best_iteration=save_iterations_filename,
            free_betas_names=function_evaluator.model_elements.expressions_registry.free_betas_names,
        )

    max_number_parameters_to_report = parameters.get('max_number_parameters_to_report')
    variable_names = (
        function_evaluator.model_elements.expressions_registry.free_betas_names
        if max_number_parameters_to_report is None
        else (
            function_evaluator.model_elements.expressions_registry.free_betas_names
            if function_evaluator.model_elements.expressions_registry.number_of_free_betas
            <= max_number_parameters_to_report
            else None
        )
    )

    return optimization(
        the_algorithm=the_algorithm,
        the_function=the_function_to_minimize,
        starting_values=function_evaluator.model_elements.expressions_registry.get_betas_array(
            starting_values
        ),
        bounds=function_evaluator.model_elements.expressions_registry.bounds,
        variable_names=variable_names,
        parameters=parameters,
    )



"""

biogeme.biogeme_logging
=======================

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Wed Nov 22 13:52:20 2023
"""

# %%
import os
import biogeme.version as ver
import biogeme.biogeme_logging as blog

# %%
# Version of Biogeme.
print(ver.getText())

# %%
# In Python, the levels of reporting are:
#
#     - DEBUG
#     - INFO
#     - WARNING
#     - ERROR
#     - CRITICAL
#
# In Biogeme, we basically use the first three.

# %%
# If we request a specific level, all message from this level and all
# levels above are displayed. For example, if INFO is requested,
# everything except DEBUG will be diplayed.
logger = blog.get_screen_logger(level=blog.INFO)


# %%
logger.info('A test')

# %%
# If a debug message is generated, it is not displayed, as the INFO
# level has been requested above.

# %%
logger.debug('A debug message')

# %%
# But a warning message is displayed, as it comes higher in the hierarchy.

# %%
logger.warning('A warning message')

# %%
# It is also possible to log the messages on file.

# %%
THE_FILE = '_test.log'

# %%
# Let's first erase the file if it happens to exist.
try:
    os.remove(THE_FILE)
    print(f'File {THE_FILE} has been erased.')
except FileNotFoundError:
    print('File {THE_FILE} does not exist.')

# %%
file_logger = blog.get_file_logger(filename=THE_FILE, level=blog.DEBUG)

# %%
file_logger.debug('A debug message')
file_logger.warning('A warning message')
file_logger.info('A info message')

# %%
# Here is the content of the log file. Note that the message includes
# the filename, which is not informative in the context of this
# Notebook.

# %%
with open(THE_FILE, encoding='utf-8') as f:
    print(f.read())



"""File spec_optima.py

:author: Michel Bierlaire, EPFL
:date: Fri Jul 29 16:01:42 2022

"""
from biogeme import models
from biogeme.expressions import Beta, Variable
import biogeme.segmentation as seg

# Definition of the variables in the data file
Id = Variable('Id')
MarginalCostPT = Variable('MarginalCostPT')
WaitingTimePT = Variable('WaitingTimePT')
CostCarCHF = Variable('CostCarCHF')
NbTransf = Variable('NbTransf')
distance_km = Variable('distance_km')
TimePT = Variable('TimePT')
TimeCar = Variable('TimeCar')
OccupStat = Variable('OccupStat')
LangCode = Variable('LangCode')
CarAvail = Variable('CarAvail')
Education = Variable('Education')
TripPurpose = Variable('TripPurpose')
Prob0 = Variable('Prob0')
Prob1 = Variable('Prob1')
Prob2 = Variable('Prob2')
Choice = Variable('Choice')

TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5

car_avail_segmentation = seg.DiscreteSegmentationTuple(
    variable=CarAvail,
    mapping={1: 'car_avail', 3: 'car_unavail'},
    reference=None,
)

language_segmentation = seg.DiscreteSegmentationTuple(
    variable=LangCode, mapping={1: 'french', 2: 'german'}, reference=None
)

occup_segmentation = seg.DiscreteSegmentationTuple(
    variable=OccupStat, mapping={1: 'full_time', 2: 'part_time', 3: 'others'}, reference=None
)

purpose_segmentation = seg.DiscreteSegmentationTuple(
    variable=TripPurpose, mapping={1: 'work', 2: 'not_work'}, reference=None
)

education_segmentation = seg.DiscreteSegmentationTuple(
    variable=Education,
    mapping={
        3: 'vocational',
        4: 'high_school',
        6: 'higher_education',
        7: 'university',
    },
    reference=None
)

ASC_PT_base = Beta('ASC_PT', 0, None, None, 0)

ASC_PT = (
    seg.Segmentation(
        ASC_PT_base,
        [car_avail_segmentation, language_segmentation],
    ).segmented_beta()
)

ASC_CAR_base = Beta('ASC_CAR', 0, None, None, 0)
ASC_CAR = (
    seg.Segmentation(
        ASC_CAR_base,
        [car_avail_segmentation, language_segmentation],
    ).segmented_beta()
)

BETA_TIME_base = Beta('BETA_TIME', 0, None, None, 0)
BETA_TIME = (
    seg.Segmentation(
        BETA_TIME_base,
        [occup_segmentation]
    ).segmented_beta()
)

BETA_COST_PT = Beta('BETA_COST_PT', 0, None, None, 0)
BETA_COST_CAR = Beta('BETA_COST_CAR', 0, None, None, 0)

BETA_WAITING_base = Beta('BETA_WAITING', 0, None, None, 0)
BETA_WAITING = seg.Segmentation(BETA_WAITING_base, [purpose_segmentation]).segmented_beta()

BETA_DIST_base = Beta('BETA_DIST', 0, None, None, 0)
BETA_DIST = seg.Segmentation(BETA_DIST_base, [education_segmentation]).segmented_beta()

LAMBDA_COST = 0.3214999879822265

V_PT = (
    ASC_PT
    + BETA_TIME * TimePT_scaled
    + BETA_COST_PT * models.boxcox(MarginalCostPT_scaled, LAMBDA_COST)
    + BETA_WAITING * WaitingTimePT**0.5
)

V_CAR = (
    ASC_CAR
    + BETA_TIME * TimeCar_scaled
    + BETA_COST_CAR * models.boxcox(CostCarCHF_scaled, LAMBDA_COST)
)

V_SM = BETA_DIST * distance_km_scaled


V = {0: V_PT, 1: V_CAR, 2: V_SM}

av = {0: 1, 1: CarAvail != 3, 2: 1}



"""

Nested logit with corrections for endogeneous sampling
======================================================

The sample is said to be endogenous if the probability for an
individual to be in the sample depends on the choice that has been
made. In that case, the ESML estimator is not appropriate anymore, and
corrections need to be made. See `Bierlaire, bolduc, McFadden (2008)
<https://dx.doi.org/10.1016/j.trb.2007.09.003>`_.

This is illustrated in this example.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:25:03 2023

"""

import numpy as np
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta
from biogeme.nests import OneNestForNestedLogit, NestsForNestedLogit

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b14nested_endogenous_sampling.py')

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
MU = Beta('MU', 1, 1, 10, 0)

# %%
# In this example, we assume that the three modes exist, and that the
# sampling protocol is choice-based. The probability that a respondent
# belongs to the sample is R_i.
R_TRAIN = 4.42e-2
R_SM = 3.36e-3
R_CAR = 7.5e-3

# %%
# The correction terms are the log of these quantities
correction = {1: np.log(R_TRAIN), 2: np.log(R_SM), 3: np.log(R_CAR)}

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of nests. Only the non trivial nests must be defined. A
# trivial nest is a nest containing exactly one alternative.  In this
# example, we create a nest for the existing modes, that is train (1)
# and car (3).

existing = OneNestForNestedLogit(
    nest_param=MU, list_of_alternatives=[1, 3], name='existing'
)

nests = NestsForNestedLogit(choice_set=list(V), tuple_of_nests=(existing,))

# %%
# The choice model is a nested logit, with corrections for endogenous sampling
# We first obtain the expression of the Gi function for nested logit.
Gi = models.getMevForNested(V, av, nests)

# %%
# Then we calculate the MEV log probability, accounting for the correction.
logprob = models.logmev_endogenousSampling(V, Gi, av, correction, CHOICE)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b14nested_endogenous_eampling'

# %%
# Estimate the parameters.
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
pandas_results = results.getEstimatedParameters()
pandas_results



"""

biogeme.loglikelihood
=====================

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Wed Nov 22 15:16:56 2023
"""
# %%
import numpy as np
import pandas as pd
import biogeme.version as ver
import biogeme.biogeme_logging as blog
import biogeme.database as db
import biogeme.loglikelihood as ll
import biogeme.expressions as ex
import biogeme.models as md

# %%
# Version of Biogeme.
print(ver.getText())

logger = blog.get_screen_logger(level=blog.INFO)

# %%
# This module provides some basic expressions for the contribution of
# an observation to the (log) likelihood function.

# %%
# Let's consider first a simple choice model.

# %%
V1 = 0
beta = ex.Beta('beta', 0, None, None, 0)
sigma = ex.Beta('sigma', 1, 0, None, 0)
V2 = beta + sigma * ex.bioDraws('V2', 'NORMAL')
V = {1: V1, 2: V2}
condprob = md.logit(V, None, 0)
prob = ex.MonteCarlo(condprob)
print(prob)

# %%
# The first function simply takes the log of the probability for each
# observation.

# %%
loglike = ll.loglikelihood(prob)
print(loglike)

# %%
# The second function also involves the integral using Monte-Carlo
# simulation.

# %%
loglike = ll.mixedloglikelihood(condprob)
print(loglike)

# %%
# Regression models are often used in the context of hybrid choice
# models. Consider the following model.

# %%
x = ex.Variable('x')
y = ex.Variable('y')
beta = ex.Beta('beta', 1, None, None, 0)
sigma = ex.Beta('sigma', 1, None, None, 0)
intercept = ex.Beta('intercept', 0, None, None, 0)
model = intercept + beta * x

# %%
# The following function calculates the contribution to the likelihood. It is
#
#  .. math:: \frac{1}{\sigma} \phi\left( \frac{y-m}{\sigma} \right),
#
#  where :math:`\phi(\cdot)` is the pdf of the normal distribution.
like = ll.likelihoodregression(y, model, sigma)
print(like)

# %%
# The following function calculates the log of the contribution to the
# likelihood. It is
#
# .. math:: -\left( \frac{(y-m)^2}{2\sigma^2} \right) -
#               \log(\sigma) - \frac{1}{2}\log(2\pi).
loglike = ll.loglikelihoodregression(y, model, sigma)
print(loglike)

# %%
# We compare the two on a small database.

# %%
df = pd.DataFrame({'x': [-2, -1, 0, 1, 2], 'y': [1, 1, 1, 1, 1]})
my_data = db.Database('test', df)

# %%
lr = like.getValue_c(my_data, prepareIds=True)
lr

# %%
np.log(lr)

# %%
loglike.getValue_c(my_data, prepareIds=True)



"""

Data processing for the optima case study
=========================================

Read the data from file, and define variables.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 16:40:59 2023

"""

import pandas as pd
import biogeme.database as db
from biogeme.expressions import Variable

# %%
# Read the data
df = pd.read_csv('optima.dat', sep='\t')
database = db.Database('optima', df)

# %%
ID = Variable('ID')
DestAct = Variable('DestAct')
NbTransf = Variable('NbTransf')
TimePT = Variable('TimePT')
WalkingTimePT = Variable('WalkingTimePT')
WaitingTimePT = Variable('WaitingTimePT')
CostPT = Variable('CostPT')
CostCar = Variable('CostCar')
TimeCar = Variable('TimeCar')
NbHousehold = Variable('NbHousehold')
NbChild = Variable('NbChild')
NbCar = Variable('NbCar')
NbMoto = Variable('NbMoto')
NbBicy = Variable('NbBicy')
NbBicyChild = Variable('NbBicyChild')
NbComp = Variable('NbComp')
NbTV = Variable('NbTV')
Internet = Variable('Internet')
NewsPaperSubs = Variable('NewsPaperSubs')
NbCellPhones = Variable('NbCellPhones')
NbSmartPhone = Variable('NbSmartPhone')
HouseType = Variable('HouseType')
OwnHouse = Variable('OwnHouse')
NbRoomsHouse = Variable('NbRoomsHouse')
YearsInHouse = Variable('YearsInHouse')
Income = Variable('Income')
Gender = Variable('Gender')
BirthYear = Variable('BirthYear')
Mothertongue = Variable('Mothertongue')
FamilSitu = Variable('FamilSitu')
OccupStat = Variable('OccupStat')
SocioProfCat = Variable('SocioProfCat')
CalculatedIncome = Variable('CalculatedIncome')
Education = Variable('Education')
HalfFareST = Variable('HalfFareST')
LineRelST = Variable('LineRelST')
GenAbST = Variable('GenAbST')
AreaRelST = Variable('AreaRelST')
OtherST = Variable('OtherST')
CarAvail = Variable('CarAvail')
MarginalCostPT = Variable('MarginalCostPT')
CostCarCHF = Variable('CostCarCHF')
Envir01 = Variable('Envir01')
Envir02 = Variable('Envir02')
Envir03 = Variable('Envir03')
Envir04 = Variable('Envir04')
Envir05 = Variable('Envir05')
Envir06 = Variable('Envir06')
Mobil01 = Variable('Mobil01')
Mobil02 = Variable('Mobil02')
Mobil03 = Variable('Mobil03')
Mobil04 = Variable('Mobil04')
Mobil05 = Variable('Mobil05')
Mobil06 = Variable('Mobil06')
Mobil07 = Variable('Mobil07')
Mobil08 = Variable('Mobil08')
Mobil09 = Variable('Mobil09')
Mobil10 = Variable('Mobil10')
Mobil11 = Variable('Mobil11')
Mobil12 = Variable('Mobil12')
Mobil13 = Variable('Mobil13')
Mobil14 = Variable('Mobil14')
Mobil15 = Variable('Mobil15')
Mobil16 = Variable('Mobil16')
Mobil17 = Variable('Mobil17')
Mobil18 = Variable('Mobil18')
Mobil19 = Variable('Mobil19')
Mobil20 = Variable('Mobil20')
Mobil21 = Variable('Mobil21')
Mobil22 = Variable('Mobil22')
Mobil23 = Variable('Mobil23')
Mobil24 = Variable('Mobil24')
Mobil25 = Variable('Mobil25')
Mobil26 = Variable('Mobil26')
Mobil27 = Variable('Mobil27')
ResidCh01 = Variable('ResidCh01')
ResidCh02 = Variable('ResidCh02')
ResidCh03 = Variable('ResidCh03')
ResidCh04 = Variable('ResidCh04')
ResidCh05 = Variable('ResidCh05')
ResidCh06 = Variable('ResidCh06')
ResidCh07 = Variable('ResidCh07')
LifSty01 = Variable('LifSty01')
LifSty02 = Variable('LifSty02')
LifSty03 = Variable('LifSty03')
LifSty04 = Variable('LifSty04')
LifSty05 = Variable('LifSty05')
LifSty06 = Variable('LifSty06')
LifSty07 = Variable('LifSty07')
LifSty08 = Variable('LifSty08')
LifSty09 = Variable('LifSty09')
LifSty10 = Variable('LifSty10')
LifSty11 = Variable('LifSty11')
LifSty12 = Variable('LifSty12')
LifSty13 = Variable('LifSty13')
LifSty14 = Variable('LifSty14')
TripPurpose = Variable('TripPurpose')
TypeCommune = Variable('TypeCommune')
UrbRur = Variable('UrbRur')
LangCode = Variable('LangCode')
ClassifCodeLine = Variable('ClassifCodeLine')
frequency = Variable('frequency')
ResidChild = Variable('ResidChild')
NbTrajects = Variable('NbTrajects')
FreqCarPar = Variable('FreqCarPar')
FreqTrainPar = Variable('FreqTrainPar')
FreqOtherPar = Variable('FreqOtherPar')
FreqTripHouseh = Variable('FreqTripHouseh')
Region = Variable('Region')
distance_km = Variable('distance_km')
Choice = Variable('Choice')
InVehicleTime = Variable('InVehicleTime')
ModeToSchool = Variable('ModeToSchool')
ReportedDuration = Variable('ReportedDuration')
CoderegionCAR = Variable('CoderegionCAR')
age = Variable('age')
Weight = Variable('Weight')

# %%
# Exclude observations such that the chosen alternative is -1
database.remove(Choice == -1.0)

# %%
# Definition of other variables
ScaledIncome = database.DefineVariable('ScaledIncome', CalculatedIncome / 1000)
age_65_more = database.DefineVariable('age_65_more', age >= 65)
moreThanOneCar = database.DefineVariable('moreThanOneCar', NbCar > 1)
moreThanOneBike = database.DefineVariable('moreThanOneBike', NbBicy > 1)
individualHouse = database.DefineVariable('individualHouse', HouseType == 1)
male = database.DefineVariable('male', Gender == 1)
haveChildren = database.DefineVariable(
    'haveChildren', ((FamilSitu == 3) + (FamilSitu == 4)) > 0
)
haveGA = database.DefineVariable('haveGA', GenAbST == 1)
highEducation = database.DefineVariable('highEducation', Education >= 6)
childCenter = database.DefineVariable(
    'childCenter', ((ResidChild == 1) + (ResidChild == 2)) > 0
)

childSuburb = database.DefineVariable(
    'childSuburb', ((ResidChild == 3) + (ResidChild == 4)) > 0
)
TimePT_scaled = database.DefineVariable('TimePT_scaled', TimePT / 200)
TimeCar_scaled = database.DefineVariable('TimeCar_scaled', TimeCar / 200)
MarginalCostPT_scaled = database.DefineVariable(
    'MarginalCostPT_scaled', MarginalCostPT / 10
)
CostCarCHF_scaled = database.DefineVariable('CostCarCHF_scaled', CostCarCHF / 10)
distance_km_scaled = database.DefineVariable('distance_km_scaled', distance_km / 5)
PurpHWH = database.DefineVariable('PurpHWH', TripPurpose == 1)
PurpOther = database.DefineVariable('PurpOther', TripPurpose != 1)



"""
DataContainer: Responsible for holding and safely manipulating
the Biogeme dataset stored as a Pandas DataFrame.

Michel Bierlaire
Wed Mar 26 19:30:57 2025
"""

from __future__ import annotations

import logging
from collections.abc import Callable

import jax.numpy as jnp
import numpy as np
import pandas as pd

from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Expression,
    ExpressionOrNumeric,
    Variable,
    validate_and_convert,
)
from biogeme.floating_point import JAX_FLOAT, PANDAS_FLOAT
from biogeme.second_derivatives import SecondDerivativesMode
from biogeme.segmentation import (
    DiscreteSegmentationTuple,
    generate_segmentation,
    verify_segmentation,
)
from .sampling import sample_with_replacement

logger = logging.getLogger(__name__)
"""Logger that controls the output of
        messages to the screen and log file.
        """


class Database:
    """Encapsulates a pandas DataFrame for Biogeme, providing safe access
    and basic operations such as checking for emptiness, scaling,
    and column manipulation.
    """

    def __init__(self, name: str, dataframe: pd.DataFrame):
        """
        Constructor

        :param name: name of the database
        :param dataframe: the data in pandas format
        :raises BiogemeError: if the dataframe is empty
        """
        self.name = name
        if dataframe.empty:
            raise BiogemeError('Database has no entry')
        try:
            self._df = dataframe.astype(PANDAS_FLOAT)
        except ValueError as e:
            raise BiogemeError(f'Data type conversion failed: {e}')
        self.number_of_excluded_data = 0

        self._listeners = []  # Called when the database is updated

        self.panel_column: str | None = None

    @classmethod
    def dummy_database(
        cls,
    ) -> Database:
        df = pd.DataFrame({'x': [0]})  # single-row dummy input
        return Database('dummy', df)

    def __str__(self) -> str:
        return f'biogeme database {self.name}'

    def register_listener(self, callback: Callable[[pd.Index], None]):
        self._listeners.append(callback)

    @property
    def dataframe(self) -> pd.DataFrame:
        """Returns a reference to the internal DataFrame."""
        return self._df

    def bootstrap_sample(self):
        """Returns a bootstrap sample of the data."""
        df = sample_with_replacement(self._df)
        return Database(f'{self.name}_bootstrap', df)

    @property
    def data_jax(self) -> jnp.ndarray:
        """Returns the data as a biogeme_jax object"""
        return jnp.asarray(self._df.to_numpy(), dtype=JAX_FLOAT)

    def is_empty(self) -> bool:
        """Returns True if the data container is empty"""
        return self._df.empty

    def num_rows(self) -> int:
        """Returns the number of rows in the dataset"""
        return self._df.shape[0]

    def num_columns(self) -> int:
        """Returns the number of columns in the dataset"""
        return self._df.shape[1]

    def column_exists(self, column: str) -> bool:
        """Check if a column exists in the data"""
        return column in self._df.columns

    def scale_column(self, column: str, scale: float):
        """Scales all values in a given column

        :param column: name of the column to scale
        :param scale: scalar to multiply the column values by
        :raises BiogemeError: if the column is not found
        """
        if column not in self._df:
            raise BiogemeError(f'Column {column} not found in the database.')
        self._df[column] *= scale

    def add_column(self, column: str, values: pd.Series) -> None:
        """Adds a new column to the dataset

        :param column: name of the new column
        :param values: a pandas Series of same length as data
        :raises ValueError: if column already exists or lengths mismatch
        """
        if column in self._df.columns:
            raise ValueError(f'Column "{column}" already exists.')

        if len(values) != self.num_rows():
            raise ValueError(
                f'Length mismatch: column has {len(values)} values, '
                f'expected {self.num_rows()}.'
            )

        self._df[column] = values

    def remove_rows(self, condition: pd.Series):
        """Removes all rows where the condition is True

        :param condition: Boolean Series of same length as the data
        """

        self._df = self._df[~condition].reset_index(drop=True)
        condition_index = condition[condition].index
        for callback in self._listeners:
            callback(condition_index)

    def reset_indices(self) -> None:
        self._df = self._df.reset_index(drop=True)

    def remove(self, exclude_condition: ExpressionOrNumeric):
        """
        Removes rows from the database that satisfy a given condition.

        This method evaluates a Biogeme expression row by row on the database.
        All rows where the expression evaluates to a truthy value are removed.

        :param exclude_condition: A Biogeme expression that returns a boolean-like value
                                  for each row in the dataset. Rows where the result is
                                  True (nonzero) will be excluded.
        """
        from biogeme.calculator import evaluate_simple_expression_per_row

        exclude_condition: Expression = validate_and_convert(exclude_condition)
        condition = evaluate_simple_expression_per_row(
            expression=exclude_condition,
            database=self,
            numerically_safe=True,
            second_derivatives_mode=SecondDerivativesMode.NEVER,
        )
        series = pd.Series(condition != 0.0)
        self.number_of_excluded_data = int(series.sum())
        self.remove_rows(series)

    def define_variable(self, name: str, expression: Expression) -> Variable:
        """
        This method evaluates a Biogeme expression row by row on the database
        and creates a new column in the internal DataFrame with the results.

        :param name: Name of the new column to be added.
        :param expression: Biogeme expression to evaluate for each row.
        """
        if name in self.dataframe.columns:
            error_msg = f'Variable {name} already exists'
            raise ValueError(error_msg)
        if self.dataframe.empty:
            error_msg = 'Empty database.'
            raise BiogemeError(error_msg)

        from biogeme.calculator import evaluate_simple_expression_per_row

        new_values = evaluate_simple_expression_per_row(
            expression=expression,
            database=self,
            numerically_safe=True,
            second_derivatives_mode=SecondDerivativesMode.NEVER,
        )
        if np.isnan(new_values).any():
            num_total = len(new_values)
            num_nan = np.isnan(new_values).sum()
            nan_indices = np.where(np.isnan(new_values))[0].tolist()

            message = f"The evaluated values for '{name}' contain NaN entries.\n"
            message += f"Total values: {num_total}, NaN values: {num_nan}.\n"

            if num_nan == num_total:
                message += "All values are NaN."
            else:
                message += f"Indices with NaN: {nan_indices}"

            raise BiogemeError(message)
        self.dataframe[name] = pd.Series(
            new_values, index=self.dataframe.index, dtype=PANDAS_FLOAT
        )
        return Variable(name)

    def remove_column(self, column: str):
        """Removes a column from the dataset"""
        if column in self._df.columns:
            self._df.drop(columns=[column], inplace=True)

    def get_column(self, column: str) -> pd.Series:
        """Returns the values of a column"""
        if column not in self._df.columns:
            raise BiogemeError(f'Column "{column}" not found.')
        return self._df[column]

    def generate_segmentation(
        self,
        variable: Variable | str,
        mapping: dict[int, str] | None = None,
        reference: str | None = None,
    ) -> DiscreteSegmentationTuple:
        """Generate a segmentation tuple for a variable.

        :param variable: Variable object or name of the variable
        :param mapping: mapping associating values of the variable to
            names. If incomplete, default names are provided.
        :param reference: name of the reference category. If None, an
            arbitrary category is selected as reference.

        """
        return generate_segmentation(
            dataframe=self.dataframe,
            variable=variable,
            mapping=mapping,
            reference=reference,
        )

    def panel(self, column_name: str):
        self.panel_column = column_name

    def verify_segmentation(self, segmentation: DiscreteSegmentationTuple) -> None:
        """Verifies if the definition of the segmentation is consistent with the data

        :param segmentation: definition of the segmentation
        :raise BiogemeError: if the segmentation is not consistent with the data.
        """
        verify_segmentation(dataframe=self.dataframe, segmentation=segmentation)

    def extract_slice(self, indices: pd.Index) -> Database:
        """
        Create a new Database instance containing only a subset of the data.

        This is useful to maintain consistency across estimation and validation datasets
        by slicing the original draws array according to the provided indices.

        :param indices: The indices used to extract the subset of draws.
        :return: A new Database instance containing the sliced draws.
        """
        sliced_database = Database(
            name=f'sliced {self.name}', dataframe=self.dataframe[indices]
        )
        return sliced_database

    def suggest_scaling(
        self, columns: list[str] | None = None, report_all: bool = False
    ) -> pd.DataFrame:
        """Suggest a scaling of the variables in the database.

        For each column, :math:`\\delta` is the difference between the
        largest and the smallest value, or one if the difference is
        smaller than one. The level of magnitude is evaluated as a
        power of 10. The suggested scale is the inverse of this value.

        .. math:: s = \\frac{1}{10^{|\\log_{10} \\delta|}}

        where :math:`|x|` is the integer closest to :math:`x`.

        :param columns: list of columns to be considered.
                        If None, all of them will be considered.

        :param report_all: if False, remove entries where the suggested
            scale is 1, 0.1 or 10

        :return: A Pandas dataframe where each row contains the name
                 of the variable and the suggested scale s. Ideally,
                 the column should be multiplied by s.

        :raise BiogemeError: if a variable in ``columns`` is unknown.
        """
        if columns is None:
            columns = self.dataframe.columns
        else:
            for c in columns:
                if c not in self.dataframe:
                    error_msg = f'Variable {c} not found.'
                    raise BiogemeError(error_msg)

        largest_value = [
            max(np.abs(self.dataframe[col].max()), np.abs(self.dataframe[col].min()))
            for col in columns
        ]
        res = [
            [col, 1 / 10 ** np.round(np.log10(max(1.0, lv))), lv]
            for col, lv in zip(columns, largest_value)
        ]
        df = pd.DataFrame(res, columns=['Column', 'Scale', 'Largest'])
        if not report_all:
            # Remove entries where the suggested scale is 1, 0.1 or 10
            remove = (df.Scale == 1) | (df.Scale == 0.1) | (df.Scale == 10)
            df.drop(df[remove].index, inplace=True)
        return df

    def is_panel(self) -> bool:
        return self.panel_column is not None



"""
This module defines the Draws class, which manages the generation and
conversion of random draws for use in simulation-based models.

Michel Bierlaire
Thu Mar 27 08:42:16 2025
"""

from __future__ import annotations

import logging
from datetime import datetime, timedelta

import jax.numpy as jnp
import numpy as np
import pandas as pd

from biogeme.floating_point import JAX_FLOAT
from .factory import DrawFactory
from .native_draws import RandomNumberGeneratorTuple

LOW_NUMBER_OF_DRAWS = 1000
logger = logging.getLogger(__name__)


class DrawsManagement:
    """
    Manages generation of simulation draws and conversion to JAX-compatible format.
    """

    def __init__(
        self,
        sample_size: int,
        number_of_draws: int,
        user_generators: dict[str, RandomNumberGeneratorTuple] | None = None,
    ):
        """
        Constructor for the Draws class.

        :param sample_size: The number of observations (rows) in the sample.
        :param number_of_draws: The number of draws to generate per observation.
        :param user_generators: Optional dictionary of user-defined random number generators.
        """
        if sample_size <= 0:
            raise ValueError(f'Incorrect sample size: {sample_size}')

        self.user_generators = user_generators
        self.sample_size: int = sample_size
        self.number_of_draws: int = number_of_draws
        self.factory: DrawFactory = DrawFactory(user_generators)
        self.draws: np.ndarray | None = None
        self.draw_types: dict[str, str] | None = None
        self.processing_time: timedelta = timedelta(0)

    def generate_draws(
        self,
        draw_types: dict[str, str],
        variable_names: list[str],
    ) -> None:
        """
        Generates random draws using the configured factory.

        :param draw_types: Mapping of variable names to draw types.
        :param variable_names: List of variable names requiring draws.
        :return: The generated draws as a NumPy array.
        """
        self.draw_types = draw_types
        if self.number_of_draws <= 0:
            raise ValueError(f'Incorrect number of draws: {self.number_of_draws}')
        if self.number_of_draws <= LOW_NUMBER_OF_DRAWS:
            warning_msg = f'The number of draws ({self.number_of_draws}) is low. The results may not be meaningful.'
            logger.warning(warning_msg)
        start_time = datetime.now()
        self.draws = self.factory.generate_draws(
            draw_types=draw_types,
            variable_names=variable_names,
            sample_size=self.sample_size,
            number_of_draws=self.number_of_draws,
        )
        self.processing_time = datetime.now() - start_time

    @property
    def draws_jax(self) -> jnp.ndarray:
        """
        Returns the generated draws as a JAX array.
        If no draws have been generated, returns an empty JAX array of shape (sample_size, 1, 1).

        :return: JAX-compatible array of draws.
        """
        if self.draws is not None:
            return jnp.asarray(self.draws, dtype=JAX_FLOAT)
        return jnp.zeros((self.sample_size, 1, 1), dtype=JAX_FLOAT)

    def extract_slice(self, indices: pd.Index) -> DrawsManagement:
        """
        Create a new DrawsManagement instance containing only a subset of draws.

        This is useful to maintain consistency across estimation and validation datasets by slicing
        the original draws array according to the provided indices.

        :param indices: The indices used to extract the subset of draws.
        :return: A new DrawsManagement instance containing the sliced draws.
        """
        sliced_draw_management: DrawsManagement = DrawsManagement(
            sample_size=len(indices),
            number_of_draws=self.number_of_draws,
            user_generators=self.user_generators,
        )
        sliced_draw_management.draws = (
            self.draws[indices] if self.draws is not None else None
        )
        return sliced_draw_management

    def remove_rows(self, indices: pd.Index) -> None:
        """Remove rows. Typically called when the database is modified."""
        if self.draws is None:
            return
        self.draws = self.draws[indices]



"""

Antithetic draws
================

Calculation of a simple integral using Monte-Carlo integration. It
illustrates how to use antithetic draws.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 20:48:02 2023
"""

import numpy as np
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme import draws
from biogeme.expressions import exp, bioDraws, MonteCarlo

# %%
# We create a fake database with one entry, as it is required to store
# the draws.
df = pd.DataFrame()
df['FakeColumn'] = [1.0]
database = db.Database('fake_database', df)


# %%
def halton13_anti(sample_size: int, number_of_draws: int) -> np.array:
    """The user can define new draws. For example, antithetic Halton
    draws with base 13, skipping the first 10 draws.

    :param sample_size: number of observations for which draws must be
                       generated.
    :param number_of_draws: number of draws to generate.

    """

    # We first generate half of the number of requested draws.
    the_draws = draws.getHaltonDraws(
        sample_size, int(number_of_draws / 2.0), base=13, skip=10
    )
    # We complement them with their antithetic version.
    return np.concatenate((the_draws, 1 - the_draws), axis=1)


# %%
mydraws = {
    'HALTON13_ANTI': (
        halton13_anti,
        'Antithetic Halton draws, base 13, skipping 10',
    )
}
database.setRandomNumberGenerators(mydraws)

# %%
# Integrate with antithetic pseudo-random number generator.
integrand = exp(bioDraws('U', 'UNIFORM_ANTI'))
simulatedI = MonteCarlo(integrand)

# %%
# Integrate with antithetic Halton draws, base 13.
integrand_halton13 = exp(bioDraws('U_halton13', 'HALTON13_ANTI'))
simulatedI_halton13 = MonteCarlo(integrand_halton13)

# %%
# Integrate with antithetic MLHS.
integrand_mlhs = exp(bioDraws('U_mlhs', 'UNIFORM_MLHS_ANTI'))
simulatedI_mlhs = MonteCarlo(integrand_mlhs)

# %%
# True value
trueI = exp(1.0) - 1.0

# %%
# Number of draws.
R = 20000

# %%
error = simulatedI - trueI

# %%
error_halton13 = simulatedI_halton13 - trueI

# %%
error_mlhs = simulatedI_mlhs - trueI

# %%
simulate = {
    'Analytical Integral': trueI,
    'Simulated Integral': simulatedI,
    'Error             ': error,
    'Simulated Integral (Halton13)': simulatedI_halton13,
    'Error (Halton13)             ': error_halton13,
    'Simulated Integral (MLHS)': simulatedI_mlhs,
    'Error (MLHS)             ': error_mlhs,
}


# %%
biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b03antithetic'
results = biosim.simulate(theBetaValues={})
results

# %%
# Reorganize the results.
print(f"Analytical integral: {results.iloc[0]['Analytical Integral']:.6f}")
print(
    f"         \t{'Uniform (Anti)':>15}\t{'Halton13 (Anti)':>15}\t{'MLHS (Anti)':>15}"
)
print(
    f"Simulated\t{results.iloc[0]['Simulated Integral']:>15.6g}\t"
    f"{results.iloc[0]['Simulated Integral (Halton13)']:>15.6g}\t"
    f"{results.iloc[0]['Simulated Integral (MLHS)']:>15.6g}"
)
print(
    f"Error\t\t{results.iloc[0]['Error             ']:>15.6g}\t"
    f"{results.iloc[0]['Error (Halton13)             ']:>15.6g}\t"
    f"{results.iloc[0]['Error (MLHS)             ']:>15.6g}"
)



from .boxcox import boxcox

from .ordered import ordered_likelihood, ordered_logit, ordered_probit

from .logit import loglogit, logit

from .mev import (
    logmev,
    mev,
    logmev_endogenous_sampling,
    mev_endogenous_sampling,
    logmev_endogenousSampling,
    mev_endogenousSampling,
)
from .piecewise import (
    piecewise_variables,
    piecewise_formula,
    piecewise_as_variable,
    piecewise_function,
)

from .nested import (
    get_mev_generating_for_nested,
    get_mev_for_nested,
    get_mev_for_nested_mu,
    nested,
    lognested,
    nested_mev_mu,
    lognested_mev_mu,
    lognestedMevMu,
    getMevForNestedMu,
    getMevGeneratingForNested,
    getMevForNested,
    nestedMevMu,
)

from .cnl import (
    cnl_avail,
    logcnl_avail,
    get_mev_for_cross_nested,
    get_mev_for_cross_nested_mu,
    getMevForCrossNested,
    cnl,
    logcnl,
    cnlmu,
    logcnlmu,
    getMevForCrossNestedMu,
)



import unittest
import numpy as np

from biogeme.function_output import (
    convert_to_dict,
    FunctionOutput,
    NamedFunctionOutput,
    DisaggregateFunctionOutput,
)


class TestConvertToDict(unittest.TestCase):
    def test_normal_conversion(self):
        sequence = ['apple', 'banana', 'cherry']
        the_map = {'first': 0, 'second': 1, 'third': 2}
        expected = {'first': 'apple', 'second': 'banana', 'third': 'cherry'}
        self.assertEqual(convert_to_dict(sequence, the_map), expected)

    def test_index_out_of_bounds(self):
        sequence = ['apple', 'banana', 'cherry']
        the_map = {'first': 0, 'third': 10}
        with self.assertRaises(IndexError):
            convert_to_dict(sequence, the_map)

    def test_empty_inputs(self):
        self.assertEqual(convert_to_dict([], {}), {})


class TestFunctionOutput(unittest.TestCase):
    def test_initialization(self):
        output = FunctionOutput(2.0, np.array([1, 2]), np.array([[1, 0], [0, 1]]))
        self.assertEqual(output.function, 2.0)
        self.assertTrue((output.gradient == np.array([1, 2])).all())
        self.assertTrue((output.hessian == np.array([[1, 0], [0, 1]])).all())


class TestNamedFunctionOutput(unittest.TestCase):
    def test_initialization(self):
        f_output = FunctionOutput(3.5, np.array([0.5, 1.5]), np.array([[2, 0], [0, 3]]))
        mapping = {'zero': 0, 'one': 1}
        named_output = NamedFunctionOutput(function_output=f_output, mapping=mapping)
        expected_gradient = {'zero': 0.5, 'one': 1.5}
        expected_hessian = {'zero': {'zero': 2, 'one': 0}, 'one': {'zero': 0, 'one': 3}}
        self.assertEqual(named_output.function, 3.5)
        self.assertEqual(named_output.gradient, expected_gradient)
        self.assertEqual(named_output.hessian, expected_hessian)

    def test_with_none_gradients_and_hessians(self):
        f_output = FunctionOutput(1.0)
        mapping = {'zero': 0, 'one': 1}
        named_output = NamedFunctionOutput(function_output=f_output, mapping=mapping)
        self.assertIsNone(named_output.gradient)
        self.assertIsNone(named_output.hessian)


class TestNamedFunctionOutputFromDisaggregate(unittest.TestCase):
    def test_initialization(self):
        output = DisaggregateFunctionOutput(
            functions=np.array([3.0]),
            gradients=np.array([[1.0, 1.5]]),
            hessians=np.array([[[1.0, 0.0], [0.0, 1.0]]]),
            bhhhs=np.array([[[2.0, 2.0], [2.0, 2.0]]]),
        )
        mapping = {'zero': 0, 'one': 1}
        unique = output.unique_entry()
        named_output = NamedFunctionOutput(function_output=unique, mapping=mapping)
        expected_bhhh = {
            'zero': {'zero': 2.0, 'one': 2.0},
            'one': {'zero': 2.0, 'one': 2.0},
        }
        self.assertEqual(named_output.bhhh, expected_bhhh)


class TestDisaggregateFunctionOutput(unittest.TestCase):
    def test_unique_entry_single(self):
        output = DisaggregateFunctionOutput(
            functions=np.array([1.0]),
            gradients=np.array([[1.0, 2.0]]),
            hessians=np.array([[[1.0, 0.0], [0.0, 1.0]]]),
            bhhhs=np.array([[[2.0, 2.0], [2.0, 2.0]]]),
        )
        unique = output.unique_entry()
        self.assertIsInstance(unique, FunctionOutput)
        self.assertEqual(unique.function, 1.0)
        self.assertTrue((unique.gradient == np.array([1.0, 2.0])).all())
        self.assertTrue((unique.hessian == np.array([[1.0, 0.0], [0.0, 1.0]])).all())
        self.assertTrue((unique.bhhh == np.array([[2.0, 2.0], [2.0, 2.0]])).all())

    def test_unique_entry_multiple(self):
        output = DisaggregateFunctionOutput(functions=np.array([1.0, 2.0]))
        self.assertIsNone(output.unique_entry())

    def test_length(self):
        output = DisaggregateFunctionOutput(functions=np.array([1.0, 2.0, 3.0]))
        self.assertEqual(len(output), 3)


if __name__ == '__main__':
    unittest.main()



"""

Catalog of nonlinear specifications
===================================

Investigate of nonlinear specifications for the travel time variables:

    - linear specification,
    - Box-Cox transform,
    - power series,

for a total of 3 specifications.
See `Bierlaire and Ortelli (2023)
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_.


:author: Michel Bierlaire, EPFL
:date: Thu Jul 13 21:31:54 2023

"""
import biogeme.biogeme as bio
import biogeme.biogeme_logging as blog
from biogeme import models
from biogeme.expressions import Expression, Beta
from biogeme.models import boxcox
from biogeme.catalog import Catalog
from biogeme.results import compile_estimation_results, pareto_optimal

# %%
# See :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, 0, 0)
B_COST = Beta('B_COST', 0, None, 0, 0)

# %%
# Non linear specifications for the travel time.

# %%
# Parameter of the Box-Cox transform.
ell_travel_time = Beta('lambda_travel_time', 1, -10, 10, 0)

# %%
# Coefficients of the power series.
square_tt_coef = Beta('square_tt_coef', 0, None, None, 0)
cube_tt_coef = Beta('cube_tt_coef', 0, None, None, 0)


# %%
# Function calculation the power series.
def power_series(the_variable: Expression) -> Expression:
    """Generate the expression of a polynomial of degree 3

    :param the_variable: variable of the polynomial
    """
    return (
        the_variable
        + square_tt_coef * the_variable**2
        + cube_tt_coef * the_variable * the_variable**3
    )


# %%
# Train travel time

# %%
# Linear specification.
linear_train_tt = TRAIN_TT_SCALED

# %%
# Box-Cox transform.
boxcox_train_tt = boxcox(TRAIN_TT_SCALED, ell_travel_time)

# %%
# Power series.
power_train_tt = power_series(TRAIN_TT_SCALED)

# %%
# Definition of the catalog.
train_tt_catalog = Catalog.from_dict(
    catalog_name='train_tt_catalog',
    dict_of_expressions={
        'linear': linear_train_tt,
        'boxcox': boxcox_train_tt,
        'power': power_train_tt,
    },
)

# %%
# Swissmetro travel time

# %%
# Linear specification.
linear_sm_tt = SM_TT_SCALED

# %%
# Box-Cox transform.
boxcox_sm_tt = boxcox(SM_TT_SCALED, ell_travel_time)

# %%
# Power series.
power_sm_tt = power_series(SM_TT_SCALED)

# %%
# Definition of the catalog. Note that the controller is the same as for train.
sm_tt_catalog = Catalog.from_dict(
    catalog_name='sm_tt_catalog',
    dict_of_expressions={
        'linear': linear_sm_tt,
        'boxcox': boxcox_sm_tt,
        'power': power_sm_tt,
    },
    controlled_by=train_tt_catalog.controlled_by,
)

# %%
# Car travel time

# %%
# Linear specification.
linear_car_tt = CAR_TT_SCALED

# %%
# Box-Cox transform.
boxcox_car_tt = boxcox(CAR_TT_SCALED, ell_travel_time)

# %%
# Power series.
power_car_tt = power_series(CAR_TT_SCALED)

# %%
# Definition of the catalog. Note that the controller is the same as for train.
car_tt_catalog = Catalog.from_dict(
    catalog_name='car_tt_catalog',
    dict_of_expressions={
        'linear': linear_car_tt,
        'boxcox': boxcox_car_tt,
        'power': power_car_tt,
    },
    controlled_by=train_tt_catalog.controlled_by,
)

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME * train_tt_catalog + B_COST * TRAIN_COST_SCALED
V2 = B_TIME * sm_tt_catalog + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * car_tt_catalog + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b02nonlinear'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# %%
# Estimate the parameters.
dict_of_results = the_biogeme.estimate_catalog()

# %%
# Number of estimated models.
print(f'A total of {len(dict_of_results)} models have been estimated')

# %%
# All estimation results
compiled_results, specs = compile_estimation_results(
    dict_of_results, use_short_names=True
)

# %%
compiled_results

# %%
# Glossary
for short_name, spec in specs.items():
    print(f'{short_name}\t{spec}')

# %%
# Estimation results of the Pareto optimal models.
pareto_results = pareto_optimal(dict_of_results)
compiled_pareto_results, pareto_specs = compile_estimation_results(
    pareto_results, use_short_names=True
)

# %%
compiled_pareto_results

# %%
# Glossary.
for short_name, spec in pareto_specs.items():
    print(f'{short_name}\t{spec}')



"""Tests for logical expressions

Michel Bierlaire
Sat Jun 21 2025, 10:57:30
"""

import unittest

from biogeme.expressions import Numeric
from biogeme.expressions.logical_and import And


class TestComparisonExpressions(unittest.TestCase):
    def setUp(self):
        self.a = Numeric(3)
        self.b = Numeric(0)
        self.c = Numeric(3)

    def test_and(self):
        expr = self.a & self.c
        self.assertIsInstance(expr, And)
        self.assertEqual(expr.get_value(), 1.0)
        expr_2 = self.a & self.b
        self.assertEqual(expr_2.get_value(), 0.0)
        expr_3 = self.b & self.a
        self.assertEqual(expr_3.get_value(), 0.0)



import unittest

import pandas as pd
from biogeme.catalog import Catalog, segmentation_catalogs
from biogeme.database import Database
from biogeme.expressions import (
    Beta,
    MonteCarlo,
    MultipleExpression,
    Variable,
    log,
)
from biogeme.models import loglogit


class TestEmbedExpression(unittest.TestCase):
    def test_embed_expression(self):
        beta_1 = Beta('beta_1', 0, None, None, 0)
        var_1 = Variable('var_1')
        expression = beta_1 * var_1
        self.assertTrue(expression.embed_expression(Beta))
        self.assertTrue(expression.embed_expression(Variable))
        self.assertFalse(expression.embed_expression(MonteCarlo))


class TestMultipleExpressions(unittest.TestCase):
    def setUp(self):
        data = {
            'choice': [0, 1, 2, 1, 0],
            'male': [0, 1, 0, 1, 0],
            'train_tt': [15.593820, 11.937003, 8.916655, 1.999498, 9.184978],
            'train_cost': [6.674172, 2.857336, 13.017769, 1.128232, 14.439975],
            'sm_tt': [18.771054, 0.015575, 19.844231, 12.349630, 12.233063],
            'sm_cost': [0.141326, 0.461249, 10.495493, 7.997219, 0.933313],
            'car_tt': [19.475110, 4.655427, 1.812129, 12.367720, 7.649240],
            'car_cost': [19.664618, 9.335258, 17.198808, 13.606151, 9.009985],
        }

        df = pd.DataFrame(data)

        self.database = Database(name='test', dataframe=df)
        asc_car = Beta('asc_car', 0, None, None, 0)
        asc_train = Beta('asc_train', 0, None, None, 0)
        b_time = Beta('b_time', 0, None, None, 0)
        b_cost = Beta('b_cost', 0, None, None, 0)

        # %%
        segmentation_gender = self.database.generate_segmentation(
            variable=Variable('male'), mapping={0: 'female', 1: 'male'}
        )

        # %%
        # We define catalogs with two different specifications for the
        # ASC_CAR: non segmented, and segmented.
        self.asc_train_catalog, self.asc_car_catalog = segmentation_catalogs(
            generic_name='asc',
            beta_parameters=[asc_train, asc_car],
            potential_segmentations=(segmentation_gender,),
            maximum_number=1,
        )

        # %%
        # We now define a catalog  with the log travel time as well as the travel time.

        # %%
        # First for train
        self.train_tt_catalog = Catalog.from_dict(
            catalog_name='train_tt_catalog',
            dict_of_expressions={
                'linear': Variable('train_tt'),
                'log': log(Variable('train_tt')),
            },
        )
        # %%
        # Then for SM. But we require that the specification is the same as
        # train by defining the same controller.
        self.sm_tt_catalog = Catalog.from_dict(
            catalog_name='sm_tt_catalog',
            dict_of_expressions={
                'linear': Variable('sm_tt'),
                'log': Variable('sm_tt'),
            },
            controlled_by=self.train_tt_catalog.controlled_by,
        )

        # %%
        # Definition of the utility functions with linear cost.
        v_train = (
            self.asc_train_catalog
            + b_time * self.train_tt_catalog
            + b_cost * Variable('train_cost')
        )
        v_swissmetro = b_time * self.sm_tt_catalog + b_cost * Variable('sm_cost')
        v_car = (
            self.asc_car_catalog
            + b_time * Variable('car_tt')
            + b_cost * Variable('car_cost')
        )

        # %%
        # Associate utility functions with the numbering of alternatives.
        v = {1: v_train, 2: v_swissmetro, 3: v_car}

        # %%
        # Definition of the model. This is the contribution of each
        # observation to the log likelihood function.
        self.log_probability = loglogit(v, None, Variable('choice'))

    def test_embed_expression(self):
        self.assertTrue(self.asc_car_catalog.embed_expression(Catalog))
        self.assertTrue(self.log_probability.embed_expression(Variable))
        self.assertTrue(self.log_probability.embed_expression(Catalog))
        self.assertTrue(self.log_probability.embed_expression(MultipleExpression))


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from biogeme import models
from biogeme.biogeme import BIOGEME
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.draws import RandomNumberGeneratorTuple
from biogeme.expressions import Beta, Draws, MonteCarlo, PanelLikelihoodTrajectory, log

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)

database.panel('ID')

#

ASC_CAR = Beta('ASC_CAR', -0.458093, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -1.110543, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', -0.710008, None, None, 0)
B_COST = Beta('B_COST', -0.988997, None, None, 0)

SIGMA_CAR = Beta('SIGMA_CAR', 0.130422, None, None, 0)
SIGMA_SM = Beta('SIGMA_SM', 0.230524, None, None, 0)
SIGMA_TRAIN = Beta('SIGMA_TRAIN', 0.168784, None, None, 0)


# Provide my own random number generator to the database.
# See the numpy.random documentation to obtain a list of other distributions.
def the_triangular_generator(sample_size, number_of_draws):
    return np.random.triangular(-1, 0, 1, (sample_size, number_of_draws))


my_random_number_generators = {
    'TRIANGULAR': RandomNumberGeneratorTuple(
        the_triangular_generator,
        'Triangular distribution T(-1,0,1)',
    )
}

# Define a random parameter, with a triangular distribution, designed
# to be used for Monte-Carlo simulation
EC_CAR = SIGMA_CAR * Draws('EC_CAR', 'TRIANGULAR')
EC_SM = SIGMA_SM * Draws('EC_SM', 'TRIANGULAR')
EC_TRAIN = SIGMA_TRAIN * Draws('EC_TRAIN', 'TRIANGULAR')


SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED + EC_TRAIN
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED + EC_SM
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED + EC_CAR

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

obs_prob = models.logit(V, av, CHOICE)
cond_prob_indiv = PanelLikelihoodTrajectory(obs_prob)
log_prob = log(MonteCarlo(cond_prob_indiv))


class test_26(unittest.TestCase):
    def testEstimation(self):
        biogeme = BIOGEME(
            database,
            log_prob,
            number_of_draws=100,
            seed=1111,
            random_number_generators=my_random_number_generators,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(
            results.final_log_likelihood, -3915.6714670664164, delta=10
        )


if __name__ == '__main__':
    unittest.main()
